"""
Exceptions.

- - - - - -
Docs: https://edzed.readthedocs.io/en/latest/
Home: https://github.com/xitop/edzed/
"""

__all__ = ['EdzedError', 'EdzedInvalidState']

class EdzedError(Exception):
    """Critical error."""

class EdzedInvalidState(EdzedError):
    """Invalid state error."""



"""Time related constants."""

__all__ = ['SEC_PER_DAY', 'SEC_PER_HOUR', 'SEC_PER_MIN', 'MONTH_NAMES']

SEC_PER_DAY = 86_400
SEC_PER_HOUR = 3_600
SEC_PER_MIN = 60
MONTH_NAMES = (
    None,
    'January', 'February', 'March', 'April', 'May', 'June',
    'July', 'August', 'September', 'October', 'November', 'December')



"""
Circuit blocks.

- - - - - -
Docs: https://edzed.readthedocs.io/en/latest/
Project home: https://github.com/xitop/edzed/
"""

import abc
import collections.abc as cabc
from dataclasses import dataclass
import difflib
import logging
from typing import Any, Callable, Iterable, Mapping, Optional, Sequence, Tuple, Union
import weakref

from .exceptions import EdzedError, EdzedInvalidState


__all__ = [
    'UNDEF', 'Const', 'Block', 'CBlock', 'SBlock',
    'Event', 'EventType', 'EventCond', 'event_tuple', 'checkname',
    ]

_logger = logging.getLogger(__package__)


class _UndefType:
    """
    Type for uninitialized circuit block's output value (UNDEF).

    UNDEF is a singleton.
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __bool__(self):
        return False

    def __repr__(self):
        return "<UNDEF>"


UNDEF = _UndefType()


class Const:
    """
    A constant value to be fed into a circuit block's input.

    Unlike other blocks, Const blocks are not subclassed from the Block
    base class and they are not stored in a Circuit object. They appear
    only as inputs of other blocks.

    Public attributes for compatibility with other blocks:
        name -- automatically created from the value
        output -- constant value
    """
    _instances = weakref.WeakValueDictionary()

    def __new__(cls, const):
        try:
            return cls._instances[const]
            # __init__ will be invoked anyway
        except KeyError:
            hashable = True
        except TypeError:
            hashable = False
        new = super().__new__(cls)
        if hashable:
            cls._instances[const] = new
        return new

    def __init__(self, const: Any):
        self._output = const

    @property
    def output(self) -> Any:
        return self._output

    @property
    def name(self) -> str:
        return str(self)

    def __str__(self):
        return f"<{type(self).__name__} {self.output!r}>"


def checkname(name, nametype):
    """Raise if name is not valid."""
    if not isinstance(name, str):
        raise TypeError(f"{nametype} must be a string, but got {name!r}")
    if not name:
        raise ValueError(f"{nametype} must be a non-empty string")


class EventType:
    """
    A base class for all special event types.

    The regular event type is str (plain string like 'put').
    """


EFilter = Callable[[Mapping], Any]
EventsArg = Union['Event', Iterable['Event'], Sequence['Event']]
EFiltersArg = Union[EFilter, Iterable[EFilter], Sequence[EFilter]]


class Event:
    """
    An event to be send.
    """

    # all circuit Events to check by Circuit._resolve_events, cleared afterward
    instances = weakref.WeakSet()

    def __init__(
            self,
            dest: Union[str, 'SBlock'],
            etype: Union[str, EventType] = 'put',
            *, efilter: EFiltersArg = ()):
        self.typecheck(etype)
        self.dest = dest
        self.etype = etype
        self._filters = efilter_tuple(efilter)
        type(self).instances.add(self)

    @staticmethod
    def typecheck(etype):
        """Raise if etype is not a valid event type value."""
        if isinstance(etype, str):
            if not etype:
                raise ValueError("event name must be a non-empty string")
        elif not isinstance(etype, EventType):
            raise TypeError(f"event type must be a string or EventType, but got {etype!r}")

    # TODO in Python3.8+ (see SBlock.event for an explanation):
    #   def send(self, source: 'Block', /, **data) -> bool:
    # pylint: disable=no-method-argument, protected-access
    def send(*args, **data) -> bool:
        """
        Apply filters and send the event to the dest block.

        Add sender block's name to the event data as 'source'.

        Return True if sent, False if rejected by a filter.
        """
        self, source = args
        if self.dest.circuit is not source.circuit:
            raise EdzedError(f"event destination {self.dest} is not in the current circuit")
        data['source'] = source.name
        for efilter in self._filters:
            retval = efilter(data)
            if isinstance(retval, dict):
                data = retval
            elif not retval:
                source.log(f"Not sending event {self} (rejected by a filter)")
                return False
        dest = self.dest
        if not dest.is_initialized():
            # a destination block may be uninitialized, because events
            # may be generated during initialization process
            dest.log("pending event, initializing early")
            source.circuit.init_sblock(dest)
        source.log("sending event %s", self)
        dest.event(self.etype, **data)
        return True

    def __str__(self):
        dest = self.dest
        # give correct result even before resolving the dest name to the dest block
        dest_name = dest if isinstance(dest, str) else dest.name
        return f"<{type(self).__name__} dest='{dest_name}', event='{self.etype}'>"


def _is_multiple(arg):
    """
    Check if arg specifies multiple ordered items (inputs, events, etc.)

    The check is based on the type. The actual item count does not
    matter and can be any value including zero or one.

    Iterators are also considered to be multiple items. Remember that
    they can be iterated over only once.

    The str type arg is considered as a single name and not as a string
    of multiple characters. That's why the return value is False.

    A set is unordered and thus not recognized as multiple items by
    this function. Return value is False.

    """
    return not isinstance(arg, str) and isinstance(arg, (cabc.Sequence, cabc.Iterator))


def _to_tuple(args, validator):
    """
    Transform 'args' to a tuple of items. Validate each item.

    The validation is deemed successful unless the validator raises.
    """
    if isinstance(args, tuple):
        pass
    elif _is_multiple(args):
        args = tuple(args)
    else:
        args = (args,)
    for arg in args:
        validator(arg)
    return args


def event_tuple(events: EventsArg) -> Tuple[Event, ...]:
    """
    Transform the argument to a tuple of events.

    Accept a single event or a sequence of events.
    """
    def validator(event):
        if not hasattr(event, 'send'):
            raise TypeError(f"Expected was an Event-like object, got {event!r}")

    return _to_tuple(events, validator)


def efilter_tuple(efilters: EFiltersArg) -> Tuple[EFilter, ...]:
    """
    Transform the argument to a tuple of event filters.

    Accept a single event filter or a sequence of filters.
    """
    def validator(efilter):
        if not callable(efilter):
            raise TypeError(f"Expected was a callable, got {efilter!r}")

    return _to_tuple(efilters, validator)


class Block:
    """
    Base class for a circuit building block.
    """
    def __init__(
            self,
            name: Optional[str], *,
            desc: str = "", on_output: EventsArg = (), _reserved=False, **kwargs):
        """
        Create a block. Add it to the circuit.
        """
        if name is None:
            name = "_@" + hex(id(self))[2:]  # id() is unique
        else:
            checkname(name, "block name")
            if name.startswith('_') and not _reserved:
                raise ValueError(f"{name!r} is a reserved name (starting with an underscore")
        self.name = name
        is_cblock = isinstance(self, CBlock)
        is_sblock = isinstance(self, SBlock)
        if not is_sblock and not is_cblock:
            raise TypeError("Can't instantiate abstract Block class")
        assert not (is_sblock and is_cblock)
        for key, value in kwargs.items():
            if not key.startswith('x_') and not key.startswith('X_'):
                raise TypeError(
                    f"'{key}' is an invalid keyword argument for {type(self).__name__}()")
            setattr(self, key, value)
        self.desc = desc
        self.debug = False
        self._output_events = event_tuple(on_output)
        # oconnections will be populated by Circuit._init_connections:
        self.oconnections = set()   # output is connected to these blocks
        self._output = UNDEF
        self.circuit = simulator.get_circuit()
        self.circuit.addblock(self)

    def is_initialized(self) -> bool:
        """Return True if the output has been initialized."""
        return self._output is not UNDEF

    @property
    def output(self) -> Any:
        """Read-only access to the output value."""
        return self._output

    def log(self, msg: str, *args, **kwargs) -> None:
        """
        Log a message if debugging is enabled.

        Block debug messages are logged with INFO severity, because
        the debug severity is used mainly for the simulator itself.
        """
        if self.debug:
            _logger.info(f"{self}: {msg}", *args, **kwargs)

    def warn(self, msg: str, *args, **kwargs) -> None:
        """Log a warning message."""
        _logger.warning(f"{self}: {msg}", *args, **kwargs)

    def _send_output_events(self, previous, value):
        for event in self._output_events:
            event.send(self, previous=previous, value=value)

    def start(self) -> None:
        """
        Pre-simulation hook.
        """

    def stop(self) -> None:
        """
        Post-simulation hook.
        """

    def has_method(self, method: str) -> bool:
        """Check if a method is defined and is not a dummy."""
        try:
            method = getattr(self, method)
        except AttributeError:
            return False
        cls = type(self)
        if method is cls.dummy_method or method is cls.dummy_async_method:
            return False
        return callable(method)

    def dummy_method(self, *args, **kwargs):
        """
        A placeholder for all optional methods.

        Key properties:
            1. it can be safely called as super().method(...)
               from any overriden method.
            2. has_method() returns False for this method
        """

    async def dummy_async_method(self, *args, **kwargs):
        """A dummy async method, see dummy_method."""

    def get_conf(self) -> Mapping[str, Any]:
        """Return static block information as a dict."""
        return {
            'class': type(self).__name__,
            'debug': self.debug,
            'desc': self.desc,
            'name': self.name,
            }

    def __str__(self):
        try:
            return f"<{type(self).__name__} '{self.name}'>"
        except AttributeError:
            # A fallback for the case the self.name was not set yet, because repr()
            # and str() are supposed to always succeed.
            # The name is not set only in a derived class' __init__() before block.__init__().
            # Try to avoid repr() and str() there.
            return super().__str__()


class Addon:
    """
    Base class for all SBlock add-ons.
    """


class CBlock(Block, metaclass=abc.ABCMeta):
    """
    Base class for combinational blocks.
    """

    class InputGetter:
        """
        A helper class for a convenient access to input values.

        An input value can be retrieved using the name as a key
        or as an attribute.
        """

        def __init__(self, blk):
            self._blk = blk

        def __getitem__(self, name):
            iblk = self._blk.inputs[name]
            if isinstance(iblk, tuple):
                return tuple(b.output for b in iblk)
            return iblk.output

        def __getattr__(self, name):
            try:
                return self[name]
            except KeyError:
                raise AttributeError(f"{self._blk} has no input '{name}'") from None

    def __init_subclass__(cls, *args, **kwargs):
        """Verify that no SBlock add-ons were added to a CBlock."""
        if issubclass(cls, Addon):
            # https://bugs.python.org/issue38085
            # raise TypeError(
            _logger.error(
                f"{cls.__name__}: SBlock add-ons are not compatible with a CBlock")
        super().__init_subclass__(*args, **kwargs)

    def __init__(self, *args, **kwargs):
        self.iconnections = set()   # will be populated by Circuit._init_connections()
        self.inputs = {}    # {"name": input(s)} where input is either:
                            #   - a block providing the input value, or
                            #   - a tuple of blocks in case of an input group
                            # When building a circuit, input blocks may be
                            # temporarily represented by their names
        self._in = self.InputGetter(self)   # _in.name and _in[name] are two ways of getting
                                            # the value of the input or input group 'name'
        super().__init__(*args, **kwargs)

    def connect(self, *args, **kwargs):
        """
        Connect inputs and input groups.

        connect() only saves the block's input connections data.
        The circuit-wide processing of interconnections will take place
        in Circuit._init_connections().
        """
        self.circuit.check_not_finalized()
        if self.inputs:
            raise EdzedInvalidState("connect() may by called only once")
        if not args and not kwargs:
            raise ValueError("No inputs to connect were given")
        if '_' in kwargs:
            raise ValueError("Input name '_' is reserved")
        # save the data for later processing by Circuit._init_connections
        if args:
            for inp in args:
                if _is_multiple(inp):
                    raise ValueError(
                        f"{inp!r} is not a single input specification; "
                        "(wrap it in Const() if it is a constant)")
            self.inputs['_'] = args
        for iname, inp in kwargs.items():
            if _is_multiple(inp):
                inp = tuple(inp)
            if inp == '':
                inp = iname
            self.inputs[iname] = inp
        return self

    def input_signature(self) -> dict:
        """
        Return a dict with so called input signature.

        The signature has the following structure:
            key = input name
            value = None, if the input is a single input, or
                    number of inputs in a group, if the input is a group
        """
        if not self.inputs:
            raise EdzedInvalidState("not connect()'ed yet")
        return {
            iname: len(ival) if isinstance(ival, tuple) else None
            for iname, ival in self.inputs.items()}

    def check_signature(self, esig: Mapping) -> dict:
        """
        Check an expected signature 'esig' with the actual one.
        """
        def setdiff_msg(actual, expected):
            """Return a message describing a diff of two sets of names."""
            unexpected = actual - expected
            missing = expected - actual
            msgparts = []
            if unexpected:
                subparts = []
                for name in unexpected:
                    suggestions = difflib.get_close_matches(name, missing, n=3)
                    if suggestions:
                        top3 = ' or '.join(repr(s) for s in suggestions)
                        subparts.append(f"{name!r} (did you mean {top3} ?)")
                    else:
                        subparts.append(repr(name))
                msgparts.append("unexpected: " + ', '.join(subparts))
            if missing:
                msgparts.append("missing: " + ', '.join(repr(name) for name in missing))
            return ", ".join(msgparts)

        def valuediff_msg(name, value, expected):
            """Return a message describing a diff in signature items."""
            if expected is None:
                if value is not None:
                    return f"{name}: is a group, expected was a single input"
            elif value is None:
                return f"{name}: is a single input, expected was a group"
            elif isinstance(expected, int):
                if value != expected:
                    return f"group {name}: input count is {value}, expected was {expected}"
            else:
                try:
                    cmin, cmax = expected
                except Exception:
                    raise ValueError(
                        f"check_signature: input {name!r}: invalid value {expected!r}"
                        ) from None
                if cmin is not None and value < cmin:
                    return f"group {name}: input count is {value}, minimum is {cmin}"
                if cmax is not None and value > cmax:
                    return f"group {name}: input count is {value}, maximum is {cmax}"
            return None # no error

        bsig = self.input_signature()   # block signature
        if bsig != esig:
            if bsig.keys() != esig.keys():
                # names differ
                errmsg = setdiff_msg(set(bsig), set(esig))
                raise ValueError(f"Not connected correctly: {errmsg}")
            # if names are OK, values must differ
            errors = [
                msg for msg in (
                    valuediff_msg(name, bsig[name], expected)
                    for name, expected in esig.items())
                if msg is not None]
            if errors:
                raise ValueError(f"Not connected correctly: {'; '.join(errors)}")
        return bsig

    @abc.abstractmethod
    def _eval(self) -> Any:
        """Compute and return the output value."""

    def eval_block(self) -> bool:
        """
        Compute new output value. Return an output change indicator.
        """
        previous = self._output
        value = self._eval()
        if value is UNDEF:
            raise ValueError("Output value must not be <UNDEF>")
        if previous == value:
            return False
        self.log("output: %s -> %s", previous, value)
        self._output = value
        self._send_output_events(previous, value)
        return True

    def get_conf(self) -> Mapping[str, Any]:
        conf = super().get_conf()
        conf['type'] = 'combinational'
        if self.circuit.is_finalized():
            conf['inputs'] = {
                iname: tuple(g.name for g in ival) if isinstance(ival, tuple) else ival.name
                for iname, ival in self.inputs.items()}
        return conf


@dataclass(frozen=True)
class EventCond(EventType):
    """
    A conditional event type.

    Roughly equivalent to:
        etype = etrue if value else efalse
    where the value is taken from the event data item ['value'].
    Missing value is evaluated as False, i.e. 'efalse' is selected.

    None value means no event.
    """
    __slots__ = ['etrue', 'efalse']
    etrue: Union[None, str, EventType]
    efalse: Union[None, str, EventType]


class SBlock(Block):
    """
    Base class for sequential blocks, i.e. blocks with internal state.
    """

    def __init_subclass__(cls, *args, **kwargs):
        """
        Verify that all add-ons precede the SBlock in the class hierarchy.

        This is important for a correct order of method calls.
        """
        super().__init_subclass__(*args, **kwargs)
        cls._ct_handlers = {}
        sblock_seen = False
        for mro in cls.__mro__:
            if sblock_seen:
                if issubclass(mro, Addon):
                    # https://bugs.python.org/issue38085
                    # raise TypeError(
                    _logger.error(
                        f"The order of {cls.__name__} base classes is incorrect: "
                        f"add-ons like {mro.__name__} must appear before SBlock")
            elif mro is SBlock:
                sblock_seen = True

            if issubclass(mro, (SBlock, Addon)):
                for method_name, method in vars(mro).items():
                    if method_name.startswith('_event_'):
                        etype = method_name[7:]     # strip '_event_'
                        if etype not in cls._ct_handlers:
                            cls._ct_handlers[etype] = method
        assert sblock_seen

    def __init__(self, *args, **kwargs):
        if self.has_method('init_from_value'):
            self.initdef = kwargs.pop('initdef', UNDEF)
        self._event_active = False      # guard against event recursion
        super().__init__(*args, **kwargs)

    def set_output(self, value: Any) -> None:
        """Set new output value."""
        if value is UNDEF:
            raise ValueError("Output value must not be <UNDEF>")
        previous = self._output
        if previous == value:
            return
        self.log("output: %s -> %s", previous, value)
        self._output = value
        self.circuit.sblock_queue.put_nowait(self)
        self._send_output_events(previous, value)

    def _event(self, etype, data):  # pylint: disable=unused-argument, no-self-use
        """
        Handle an event.
        """
        return NotImplemented

    # TODO: There is an issue fixed in Python3.8+ by PEP570, but
    # we want to support Python 3.7, so a workaround must be used.
    #
    # Python 3.8 function definition will be:
    #   def event(self, etype: Union[str, EventType], /, **data) -> Any:
    #
    # The problem with:
    #   def event(self, etype, **data):     # type hints removed
    # is that names used as positional arguments (self and etype in this case)
    # cannot be used as a keyword argument in a call:
    #   block.event('foo', etype=1)         # TypeError
    #
    # pylint: disable=no-method-argument, protected-access
    def event(*args, **data) -> Any:
        """
        Error handling wrapper, see _event for function usage.

        Evaluate so called conditional events. See the EventCond class.

        Respond with ValueError to unknown event types.

        Handle the received event ETYPE:
            - invoke the specialzed _event_ETYPE() method if such
              method exists; otherwise
            - invoke the general _event() method

        IMPORTANT: application code should check Circuit.is_ready()
        before submitting external events for processing. The event()
        function must process events also during cleanup, when the
        simulator is no longer ready, but cannot distinguish internal
        and external events.
        """
        self, etype = args
        Event.typecheck(etype)
        if data:
            self.log("got event %r, data: %s", etype, data)
        else:
            self.log("got event %r", etype)
        if self._event_active:
            raise EdzedError(f"{self}: Forbidden recursive event() call")
        self._event_active = True
        try:
            while isinstance(etype, EventCond):
                etype = etype.etrue if data.get('value') else etype.efalse
                self.log("conditional event -> %r", etype)
                if etype is None:
                    return None
            handler = type(self)._ct_handlers.get(etype)    # pylint: disable=protected-access
            try:
                # handler is an unbound method, bind it with handler.__get__(self)
                retval = handler.__get__(self)(**data) if handler else self._event(etype, data)
            except Exception as err:
                if not self.circuit.is_current_task() and err.__traceback__.tb_next is not None:
                    # 1. The raised exception won't be handled by the simulation task.
                    # 2. The traceback has more than one level only, i.e. the handler function
                    # call itself succeded. Errors like missing arguments are thus ruled out
                    # and the error must have occurred inside the handler. The internal state
                    # of the block could have been corrupted. That's a sufficient reason for
                    # aborting the simulation.
                    sim_err = EdzedError(
                        f"{self}: {type(err).__name__} during handling of event "
                        f"'{etype}', data: {data}: {err}")
                    sim_err.__cause__ = err
                    self.circuit.abort(sim_err)
                raise
            if retval is NotImplemented:
                raise ValueError(f"{self}: Unknown event type '{etype}'")
            return retval
        finally:
            self._event_active = False

    @property
    class _enable_event:
        """
        A context manager temporarily enabling recursive events.

        Usage:
            ... during handling of an event ...
            with self._enable_event:
                self.event(...) # without _enable_event this would
                                # raise "Forbidden recursive event()"
        """
        def __init__(self, block: 'SBlock'):
            self._block = block
            self._event_saved = None

        # pylint: disable=protected-access
        def __enter__(self):
            block = self._block
            self._event_saved = block._event_active
            block._event_active = False
            return block

        def __exit__(self, *exc_info):
            self._block._event_active = self._event_saved

    def put(self, value: Any, **data) -> Any:
        """put(x) is a shortcut for event('put', value=x)."""
        return self.event('put', **data, value=value)

    def get_state(self) -> Any:
        """
        Return the internal state.

        The default implementation assumes the state is equal output.
        Must be redefined for more complex SBlocks.
        """
        return self._output

    def init_regular(self) -> None:
        """
        Initialize the internal state and the output.
        """

    # optional methods
    init_async = Block.dummy_async_method
    stop_async = Block.dummy_async_method
    init_from_value = Block.dummy_method

    def get_conf(self) -> Mapping[str, Any]:
        conf = super().get_conf()
        conf['type'] = 'sequential'
        return conf


# importing at the end when all names are defined resolves a circular import issue
# pylint: disable=wrong-import-position
from . import simulator



"""
Combinational blocks for general use.

- - - - - -
Docs: https://edzed.readthedocs.io/en/latest/
Home: https://github.com/xitop/edzed/
"""

import inspect

from .. import block

__all__ = ['Invert', 'FuncBlock', 'Compare', 'Override']


class Invert(block.CBlock):
    """
    Boolean negation.
    """
    def _eval(self):
        return not self._in['_'][0]

    def start(self):
        super().start()
        self.check_signature({'_': 1})


class FuncBlock(block.CBlock):
    """
    Create a combinational circuit block from a regular function.
    """

    def __init__(self, *args, func, unpack: bool = True, **kwargs):
        self._func = func
        self._unpack = unpack
        super().__init__(*args, **kwargs)

    def _eval(self):
        kwargs = {name: self._in[name] for name in self.inputs}
        args = kwargs.pop('_', ())
        if self._unpack:
            return self._func(*args, **kwargs)
        return self._func(args, **kwargs)

    def start(self):
        try:
            func = self._func
            self._func = inspect.signature(func).bind
            self._eval()
        except TypeError as err:
            raise TypeError(
                f"function {func.__qualname__} does not match the connected inputs: {err}")
        finally:
            self._func = func
        super().start()


class Compare(block.CBlock):
    """
    A comparator with hysteresis.
    """

    def __init__(self, *args, low, high, **kwargs):
        if high < low:
            raise ValueError("high threshold cannot be lower than low threshold")
        self._low = low
        self._high = high
        super().__init__(*args, **kwargs)

    def _eval(self):
        if self._output is block.UNDEF:
            thr = (self._low + self._high) / 2
        else:
            thr = self._low if self._output else self._high
        return self._in['_'][0] >= thr

    def start(self):
        super().start()
        self.check_signature({'_': 1})


class Override(block.CBlock):
    """
    Either pass input to output unchanged or override it with value.
    """
    def __init__(self, *args, null_value=None, **kwargs):
        self._null = null_value
        super().__init__(*args, **kwargs)

    def _eval(self):
        override = self._in.override
        return self._in.input if override == self._null else override

    def start(self):
        super().start()
        self.check_signature({'input': None, 'override': None})



"""
Event driven finite-state machine (FSM) extended with optional timers.

- - - - - -
Docs: https://edzed.readthedocs.io/en/latest/
Home: https://github.com/xitop/edzed/
"""

import asyncio
import collections
import collections.abc as cabc
import contextvars
from dataclasses import dataclass
import time
import types
from typing import Optional, Sequence, Tuple, Union

from . import addons
from . import block
from .exceptions import EdzedError
from .utils import looptimes
from .utils import timeunits


__all__ = ['convert_duration', 'fsm_event_data', 'FSM', 'Goto', 'INF_TIME']

INF_TIME = float('+inf')
fsm_event_data = contextvars.ContextVar('fsm_event_data')


def convert_duration(duration: Union[None, int, float, str]) -> Optional[float]:
    """Convenience wrapper for timeunits.convert()."""
    if duration is None:
        return None
    if isinstance(duration, int):
        duration = float(duration)
    if isinstance(duration, float):
        return max(0.0, duration)
    if isinstance(duration, str):
        return timeunits.convert(duration)
    raise ValueError(f"Invalid duration value: {duration!r}")


@dataclass(frozen=True)
class Goto(block.EventType):
    """A special event causing a direct state transition."""
    __slots__ = ['str']
    state: str


class FSM(addons.AddonPersistence, block.SBlock):
    """
    A base class for a Finite-state Machine with optional timers.
    """

    # subclasses should define:
    STATES = ()
    TIMERS = {}
    EVENTS = ()

    @classmethod
    def _check_state(cls, state):
        if state not in cls._ct_states:
            raise ValueError(f"Unknown state '{state}'")

    @classmethod
    def _build_tables(cls):
        """
        Build control tables from STATES, TIMERS and EVENTS.

        The original tables are left unchanged.
        """
        def add_transition(event, from_state, next_state):
            if from_state is not None:
                cls._check_state(from_state)
            key = (event, from_state)
            if key in cls._ct_transition:
                raise ValueError(
                    "Multiple transition definition for "
                    f"event {event!r} in state {from_state!r}")
            cls._ct_transition[key] = next_state

        # control tables must be created for each subclass
        # '_ct' (control table) prefix chosen to make a name clash unlikely
        cls._ct_states = set(cls.STATES).union(cls.TIMERS)  # all known states
        cls._ct_events = set()      # all known events
        cls._ct_transition = {}     # {(event, state): next_state}
        cls._ct_default_duration = {}   # {timed_state: duration_in_seconds_or_None}
        cls._ct_timed_event = {}        # {timed_state: event_on_timer_expiration}
        cls._ct_methods = {
            'enter': {},            # {state: enter_cb}
            'exit': {},             # {state: exit_cb}
            'cond': {},             # {event: cond_cb}
            }

        # prepare data for keyword argument parsing in __init__
        cls._ct_prefixes = [
            # prefix, prefix length, valid names (i.e. without the prefix) to check
            ('t_', 2, cls._ct_default_duration),
            ('cond_', 5, cls._ct_events),
            ('enter_', 6, cls._ct_states),
            ('exit_', 5, cls._ct_states),
            ('on_enter_', 9, cls._ct_states),
            ('on_exit_', 8, cls._ct_states),
            ]

        if isinstance(cls.STATES, str):
            # catch a frequent error: 'x' or ('x') instead of ('x',)
            # Note that this condition does not mean a single
            # state FSM - other states may be timed.
            raise ValueError(
                f"STATES: expected is a sequence of strings, did you mean: [{cls.STATES}] ?")

        if cls.STATES:
            cls._ct_default_state = cls.STATES[0]
        else:
            if not cls.TIMERS:
                raise ValueError("Cannot create a state machine with no states")
            # the TIMERS is a dict and dicts are ordered since Python 3.7
            cls._ct_default_state = next(iter(cls.TIMERS))  # the first key

        for state in cls._ct_states:
            block.checkname(state, "FSM state name")
        cls._ct_chainlimit = 3 * len(cls._ct_states)    # limit for chained transitions

        for event, from_states, next_state in cls.EVENTS:
            block.checkname(event, "FSM event name")
            if event in cls._ct_handlers:
                raise ValueError(
                    f"Ambiguous event '{event}': "
                    "the name is used for both FSM and SBlock event")
            cls._ct_events.add(event)
            if next_state is not None:
                cls._check_state(next_state)
            if from_states is None or isinstance(from_states, str):
                add_transition(event, from_states, next_state)
            else:
                for fstate in from_states:
                    add_transition(event, fstate, next_state)

        for state, (duration, event) in cls.TIMERS.items():
            try:
                cls._ct_default_duration[state] = convert_duration(duration)
            except ValueError as err:
                raise ValueError(f"TIMERS['{state}']: {err}") from None
            if isinstance(event, Goto):
                cls._check_state(event.state)
            elif event not in cls._ct_events:
                raise ValueError(
                    f"TIMERS['{state}']: undefined event '{event}'")
            cls._ct_timed_event[state] = event

        for method_name, method in vars(cls).items():
            try:
                cb_type, name = method_name.split('_', 1)
                cb_dict = cls._ct_methods[cb_type]
            except (ValueError, KeyError):
                continue
            valid_names = cls._ct_events if cb_type == 'cond' else cls._ct_states
            if name in valid_names and callable(method):
                cb_dict[name] = method

    # TODO: https://bugs.python.org/issue38085
    def __init_subclass__(cls, *args, **kwargs):
        """
        Build control tables.
        """
        # call super().__init_subclass__ first, we will later check for possible
        # event name clashes with the _ct_handlers
        super().__init_subclass__(*args, **kwargs)

        try:
            cls._build_tables()
        except Exception as err:
            fmt = f"Cannot create FSM subclass {cls.__name__}: {{}}"
            err.args = (fmt.format(err.args[0] if err.args else "<NO ARGS>"), *err.args[1:])
            raise

    def __init__(self, *args, on_notrans=(), **kwargs):
        """
        Create FSM.

        Keyword arguments named t_STATE (where STATE is a timed state)
        override the default durations of corresponding timers for this
        single instance.

        Keyword arguments named on_enter_STATE and on_exit_STATE (where
        STATE is a known state) define events to be sent when the STATE
        is entered or exited.

        Keyword arguments named cond_EVENT (where EVENT is a known
        event) define functions with the same semantics as cond_EVENT
        methods.

        Keyword argument on_notrans defines events to be sent when
        a transition is not defined.

        FSM initialization takes place in init_regular().
        """
        if type(self) is FSM:   # pylint: disable=unidiomatic-typecheck
            raise TypeError("Can't instantiate abstract FSM class")
        prefixed = collections.defaultdict(list)
        for arg in kwargs:
            for prefix, length, valid_names in self._ct_prefixes:
                if arg.startswith(prefix):
                    name = arg[length:]  # strip prefix
                    if name in valid_names:
                        prefixed[prefix].append((name, arg))
        self._duration = self._ct_default_duration
        for name, arg in prefixed['t_']:
            self._set_duration(name, kwargs.pop(arg))
        self._fsm_functions = {
            'cond': {name: kwargs.pop(arg) for name, arg in prefixed['cond_']},
            'enter': {name: kwargs.pop(arg) for name, arg in prefixed['enter_']},
            'exit': {name: kwargs.pop(arg) for name, arg in prefixed['exit_']},
            }
        self._state_events = {
            'on_enter': {
                name: block.event_tuple(kwargs.pop(arg))
                for name, arg in prefixed['on_enter_']},
            'on_exit': {
                name: block.event_tuple(kwargs.pop(arg))
                for name, arg in prefixed['on_exit_']},
            }

        self._on_notrans = block.event_tuple(on_notrans)
        self._state = block.UNDEF
        self._active_timer = None
        self._fsm_event_active = False
        self._next_event = None     # scheduled event in chained state transition
        kwargs.setdefault('initdef', self._ct_default_state)
        super().__init__(*args, **kwargs)

    @property
    def state(self) -> str:
        """Return the FSM state."""
        return self._state

    def get_state(self) -> Tuple[str, Optional[float]]:
        """
        Return the block's internal state.

        Internal state is a pair:
            FSM state, timer state
        The timer state is either None or an expiration time given
        as UNIX timestamp.

        Returned data can be restored with _restore_state.
        """
        timer = self._active_timer
        if timer is None or timer.cancelled():
            exp_timestamp = None
        else:
            exp_timestamp = looptimes.loop_to_unixtime(timer.when())
        return (self._state, exp_timestamp)

    def _restore_state(self, istate: Sequence) -> None:
        """
        Restore the internal state created by get_state().

        cond_STATE, enter_STATE and on_enter_STATE are not executed,
        because the state was entered already in the past. Now it is
        only restored.
        """
        state, exp_timestamp = istate
        self._check_state(state)
        if exp_timestamp is not None:
            remaining = exp_timestamp - time.time()
            if remaining <= 0.0:
                self.log("restore state: ignoring expired state")
                return
            try:
                timed_event = self._ct_timed_event[state]
            except KeyError:
                raise EdzedError(
                    f"cannot set a timer for a not timed state '{state}'") from None
            self._set_timer(remaining, timed_event)
        self._state = state
        self.log("state: <UNDEF> -> %s", state)
        output = self._eval()
        if output is not block.UNDEF:
            self.set_output(output)

    def init_from_value(self, value: str) -> None:
        """Initialize the internal state."""
        self.event(Goto(value))

    def stop(self) -> None:
        """Cleanup."""
        self._stop_timer()
        super().stop()

    def _set_duration(self, timed_state: str, value: Union[None, int, float, str]) -> None:
        """
        Set the duration of a timed_state.

        _set_duration() accepts various value formats, see the docs
        for TIMERS.

        Reset to default if value is None.
        """
        ct_duration = type(self)._ct_default_duration
        if timed_state not in ct_duration:
            raise ValueError(f"{timed_state!r} is not a timed state")
        if self._duration is ct_duration:
            self._duration = ct_duration.copy()   # copy in write
        self._duration[timed_state] = \
            ct_duration[timed_state] if value is None else convert_duration(value)

    def get_duration(self, timed_state: str) -> Optional[float]:
        """
        Return the timed state duration in seconds.

        Return None if the duration is not set. This includes the case
        when the state is not timed.
        """
        return self._duration.get(timed_state)

    def _set_timer(self, duration, timed_event):
        """Start the timer (low-level)."""
        self.log("timer: %.3fs before %s", duration, timed_event)
        self._active_timer = asyncio.get_running_loop().call_later(
            duration, self.event, timed_event)

    def _start_timer(self, duration, timed_event):
        """Start the timer."""
        if duration is not None:
            duration = convert_duration(duration)
        else:
            duration = self.get_duration(self._state)
            if duration is None:
                raise EdzedError(f"Timer duration for state '{self._state}' not set")
        if duration == INF_TIME:
            return
        if duration <= 0.0:
            self.log("timer: zero delay before %s", timed_event)
            self.event(timed_event)
            return
        self._set_timer(duration, timed_event)

    def _stop_timer(self):
        """Stop the timer, if any."""
        timer = self._active_timer
        if timer is not None:
            if not timer.cancelled():
                timer.cancel()
                # do not rely on the private attribute '_scheduled'
                if getattr(timer, '_scheduled', True):
                    self.log("timer: cancelled")
            self._active_timer = None

    def _send_events(self, trigger_type, state):
        """Send events triggered by 'on_enter_STATE' or 'on_exit_STATE'."""
        state_events = self._state_events[trigger_type]
        try:
            events = state_events[state]
        except KeyError:
            return
        trigger = f'{trigger_type}_{state}'
        for event in events:
            event.send(self, trigger=trigger)

    def _run_cb(self, cb_type, name):
        """
        Run methods and callbacks 'cond', 'enter' or 'exit'.

        Access to event data is provided through the context variable
        'fsm_event_data'.

        Try the external function and the class method. Return a list
        with 0, 1 or 2 return values depending on how many functions
        were actually defined.

        Do not rely on the order of calls (the function first or
        the method first), it may change in future releases.
        """
        retvals = []
        self_cb = self._fsm_functions[cb_type]
        try:
            cb = self_cb[name]
        except KeyError:
            pass
        else:
            retvals.append(cb())
        cls_cb = self._ct_methods[cb_type]
        try:
            cb = cls_cb[name].__get__(self)      # bind the method
        except KeyError:
            pass
        else:
            retvals.append(cb())
        return retvals

    def _event_ctx(self, etype, data):
        """
        Handle event. Check validity and conditions.

        Timed states look for 'duration' key in the data. If it is
        present, the value overrides the default timer duration.

        Return value:
            True = transition accepted, either executed or scheduled
                   for execution
            False = transition rejected
        """
        if isinstance(data, cabc.MutableMapping):
            # make read-only to prevent any ugly hacks
            rodata = types.MappingProxyType(data)
        else:
            rodata = data
        fsm_event_data.set(rodata)

        if isinstance(etype, Goto):
            newstate = etype.state
            self._check_state(newstate)
        else:
            if etype not in self._ct_events:
                return NotImplemented
            try:
                newstate = self._ct_transition[(etype, self.state)]
            except KeyError:
                newstate = self._ct_transition.get((etype, None), None)
            if newstate is None:
                self.log("No transition defined for event %s in state %s", etype, self._state)
                for event in self._on_notrans:
                    event.send(self, event=etype, state=self._state)
                return False
            if self.is_initialized() and not all(self._run_cb('cond', etype)):
                self.log(
                    "not executing event %s (%s -> %s), condition not satisfied",
                    etype, self.state, newstate)
                return False

        if self._fsm_event_active:
            # recursive call:
            #   - event ->  enter_STATE -> new event, or
            #   - timed event with zero duration -> next event
            if self._next_event is not None:
                raise EdzedError(
                    "Forbidden event multiplication; "
                    f"Two events ({self._next_event[0]} and {etype}) were generated "
                    "while handling a single event")
            self._next_event = (etype, data, newstate)
            return True

        self._fsm_event_active = True
        try:
            if self.is_initialized():
                self._run_cb('exit', self._state)
                self._send_events('on_exit', self._state)
                self._stop_timer()
            assert self._next_event is None
            for _ in range(self._ct_chainlimit):
                if self._next_event:
                    # intermediate state: skip generated events and exit the state immediately
                    self._run_cb('exit', self._state)
                    etype, data, newstate = self._next_event
                    self._next_event = None
                self.log("state: %s -> %s (event: %s)", self._state, newstate, etype)
                self._state = newstate
                with self._enable_event:
                    self._run_cb('enter', self._state)
                if self._next_event:
                    continue
                try:
                    timed_event = self._ct_timed_event[newstate]
                except KeyError:
                    pass    # new state is not a timed state
                else:
                    with self._enable_event:
                        self._start_timer(data.get('duration'), timed_event)
                    if self._next_event:
                        continue
                break
            else:
                raise EdzedError('Chained state transition limit reached (infinite loop?)')
            output = self._eval()
            if output is not block.UNDEF:
                self.set_output(output)
            self._send_events('on_enter', self._state)
            return True
        finally:
            self._fsm_event_active = False


    def _event(self, etype, data):
        """Wrapper creating a separate context, see _event_ctx for docs."""
        return contextvars.copy_context().run(self._event_ctx, etype, data)


    def _eval(self):    # pylint: disable=no-self-use
        """
        Compute and return the output value.

        Return UNDEF to leave the output unchanged.

        The output of a FSM is often unused. For these cases
        the default _eval just returns False.
        """
        return False



"""
Sequential blocks for general use.

- - - - - -
Docs: https://edzed.readthedocs.io/en/latest/
Home: https://github.com/xitop/edzed/
"""

import asyncio
import logging

from .. import addons
from .. import block
from .. import fsm
from ..utils import shield_cancel


__all__ = ['Input', 'InputExp', 'OutputAsync', 'OutputFunc']

_logger = logging.getLogger(__package__)


class Input(addons.AddonPersistence, block.SBlock):
    """
    Input with optional value validation.
    """

    def __init__(self, *args, schema=None, check=None, allowed=None, **kwargs):
        self._schema = schema
        self._check = check
        self._allowed = None if allowed is None else frozenset(allowed)
        super().__init__(*args, **kwargs)
        if self.initdef is not block.UNDEF:
            self._validate(self.initdef)

    def init_from_value(self, value):
        self.put(value)

    def _validate(self, value):
        if self._allowed is not None and value not in self._allowed:
            raise ValueError(
                f"Value {value!r} validation error: value not among allowed values")
        if self._check is not None and not self._check(value):
            raise ValueError(f"Value {value!r} validation error: check function not passed")
        if self._schema is not None:
            try:
                value = self._schema(value)
            except Exception as err:
                raise ValueError(f"Value {value!r} schema validation error: {err}") from None
        return value

    def _event_put(self, *, value, **_data):
        try:
            value = self._validate(value)
        except ValueError as err:
            self.warn("%s", err)
            return False
        self.set_output(value)
        return True

    _restore_state = init_from_value


class InputExp(Input, fsm.FSM):
    """
    Input with an expiration time.
    """
    STATES = ['expired', 'valid']
    TIMERS = {
        'valid': (None, fsm.Goto('expired')),
        }
    EVENTS = (
        ('start', None, 'valid'),
        )

    def __init__(self, *args, duration, expired=None, **kwargs):
        super().__init__(*args, t_valid=duration, **kwargs)
        self._validate(expired)
        self._expired = expired

    def enter_expired(self):
        Input._event_put(self, value=self._expired)

    def _event_put(self, *, value, **data):
        if not Input._event_put(self, value=value):
            return False    # value not accepted by validators
        with self._enable_event:
            # value is missing in data, but we need only the optional 'duration' item
            self.event(fsm.Goto('expired') if value == self._expired else 'start', **data)
        return True

    def init_from_value(self, value):
        self.put(value)

    def _eval(self):    # pylint: disable=no-self-use
        """Stop the FSM part from setting the output."""
        return block.UNDEF

    def get_state(self):
        """Combine states of both parts."""
        return {
            'input': Input.get_state(self),
            'fsm': fsm.FSM.get_state(self),
        }

    def _restore_state(self, state):
        fsm.FSM._restore_state(self, state['fsm'])
        if self._state is not block.UNDEF:
            Input._event_put(self, value=state['input'])


_STOP_CMD = object()

class OutputAsync(addons.AddonAsync, block.SBlock):
    """
    Run a coroutine as an output task when a value arrives.
    """

    def __init__(
            self, *args,
            coro, guard_time=0.0,
            qmode=False, on_success=(), on_error=None, stop_value=block.UNDEF,
            **kwargs):
        self._on_success = block.event_tuple(on_success)
        if on_error is None:
            on_error = block.Event('_ctrl', 'error')
        self._on_error = block.event_tuple(on_error)
        self._coro = coro
        self._guard_time = guard_time
        self._qmode = bool(qmode)
        self._stop_value = stop_value
        self._ctask = None
        self._queue = None
        super().__init__(*args, **kwargs)
        if guard_time > self.stop_timeout:
            raise ValueError(
                f"guard_time {guard_time:.3f} must not exceed "
                f"stop_timeout {self.stop_timeout:.3f} seconds.")

    def _event_put(self, *, value, **_data):
        self._queue.put_nowait(value)
        self.set_output(True)

    async def _output_task_wrapper(self, value):
        self.log("output task started for value %s", value)
        if self._qmode:
            qsize = self._queue.qsize()
            if qsize > 0:
                self.warn("%d value(s) waiting in output queue", qsize)
        try:
            retval = await self._coro(value)
        except asyncio.CancelledError:
            # it is assumed that the coroutine was cancelled by the control task
            self.log("output task cancelled")
        except Exception as err:
            self.warn("output task failed for input value %r: %r", value, err)
            for ev in self._on_error:
                ev.send(self, error=err)
        else:
            self.log("output task returned value %r", retval)
            for ev in self._on_success:
                ev.send(self, value=retval)
        if self._queue.empty():
            self.set_output(False)
        if self._guard_time > 0.0:
            try:
                await shield_cancel.shield_cancel(asyncio.sleep(self._guard_time))
            except asyncio.CancelledError:
                # shield_cancel re-reaises any CancelledError when it stops shielding
                pass

    async def _control_noqmode(self):
        """
        Start an output task for values from the queue.

        Only one task may be running at a time. New values
        are processed asap, cancelling the current task if necessary.

        Stop serving after receiving the _STOP_CMD sentinel value.
        """
        task = None
        stop = False
        queue = self._queue
        while True:
            if not stop:
                value = await queue.get()
                if value is _STOP_CMD:
                    stop = True
            if task and not task.done():
                if not stop:
                    task.cancel()
                await task
            if stop:
                break
            while not queue.empty():
                new_value = queue.get_nowait()
                if new_value is _STOP_CMD:
                    stop = True
                    break
                self.warn("Dropping %r from output queue", value)
                value = new_value
            task = asyncio.create_task(self._task_wrapper(self._output_task_wrapper(value)))

    async def _control_qmode(self):
        """
        Start an output task for values from the queue.

        Only one task may be running at a time. New values wait
        in the queue.

        Stop serving after receiving the _STOP_CMD sentinel value.
        """
        while True:
            value = await self._queue.get()
            if value is _STOP_CMD:
                break
            await self._output_task_wrapper(value)

    def start(self):
        super().start()
        self._queue = asyncio.Queue()
        cfunc = self._control_qmode if self._qmode else self._control_noqmode
        self._ctask = asyncio.create_task(self._task_wrapper(cfunc()))

    def init_regular(self):
        self.set_output(False)

    def stop(self):
        if self._ctask is None:
            return  # start wasn't called
        if self._stop_value is not block.UNDEF:
            self.put(self._stop_value)
        self._queue.put_nowait(_STOP_CMD)   # this will not cancel a running output task
        super().stop()

    async def stop_async(self):
        if self._ctask is None:
            return  # start wasn't called
        try:
            await self._ctask
        except asyncio.CancelledError:
            pass
        finally:
            self._ctask = None
        await super().stop_async()


class OutputFunc(block.SBlock):
    """
    Run a function when a value arrives.
    """

    def __init__(
            self, *args, func, on_success=(), on_error=None, stop_value=block.UNDEF, **kwargs):
        self._on_success = block.event_tuple(on_success)
        if on_error is None:
            on_error = block.Event('_ctrl', 'error')
        self._on_error = block.event_tuple(on_error)
        self._func = func
        self._stop_value = stop_value
        super().__init__(*args, **kwargs)

    def _event_put(self, *, value, **_data):
        try:
            retval = self._func(value)
        except Exception as err:
            self.warn("output function failed for input value %r: %r", value, err)
            for ev in self._on_error:
                ev.send(self, error=err)
        else:
            self.log("output function returned value %r", retval)
            for ev in self._on_success:
                ev.send(self, value=retval)

    def init_regular(self):
        """Initialize the internal state."""
        self.set_output(False)

    def stop(self):
        if self._stop_value is not block.UNDEF:
            self.put(self._stop_value)
        super().stop()



"""
asyncio.shield improved.
"""

import asyncio
from typing import Any, Awaitable


async def shield_cancel(aw: Awaitable) -> Any:
    """
    Shield from cancellation while aw is awaited.

    Any pending CancelledError is raised when aw is finished.
    """
    task = asyncio.ensure_future(aw)
    cancel_exc = None
    while True:
        try:
            retval = await asyncio.shield(task)
        except asyncio.CancelledError as err:
            if task.done():
                # cancelled from within aw
                raise
            cancel_exc = err
        else:
            break
    if cancel_exc:
        raise cancel_exc
    return retval



"""
FSM blocks for general use.

- - - - - -
Docs: https://edzed.readthedocs.io/en/latest/
Home: https://github.com/xitop/edzed/
"""

from .. import fsm

__all__ = ['Timer']


class Timer(fsm.FSM):
    """
    A timer.
    """

    STATES = ('off', 'on')
    TIMERS = {
        'on': (fsm.INF_TIME, 'stop'),
        'off': (fsm.INF_TIME, 'start'),
        }
    EVENTS = (
        ('start', None, 'on'),
        ('stop', None, 'off'),
        ('toggle', 'on', 'off'),
        ('toggle', 'off', 'on'),
        )

    def __init__(self, *args, restartable=True, **kwargs):
        super().__init__(*args, **kwargs)
        self._restartable = bool(restartable)

    def cond_start(self):
        return self._restartable or self._state != 'on'

    def cond_stop(self):
        return self._restartable or self._state != 'off'

    def _eval(self):
        return self._state == 'on'



"""
Call the 'recalc' method of all registered blocks at given times of day.

Intervals given by time, date, and weekdays are implemented on top
of this low-level service.

- - - - - -
Docs: https://edzed.readthedocs.io/en/latest/
Home: https://github.com/xitop/edzed/
"""

import asyncio
import bisect
from dataclasses import dataclass
import time

from . import addons
from . import block
from .utils import tconst
from .utils import timeinterval

# HMS is an abbreviation for clock time hour, minute, second
# MD is an abbreviation for month, day
# see edzed.utils.timeinterval for details

_MAX_TRACKING_ERROR = 2.0   # max. acceptable scheduler's error in seconds, must be >= 1.0


@dataclass(frozen=True)
class TimeData:
    """Time in various represenations."""
    __slots__ = ['hms', 'tstruct', 'subsec']
    hms: timeinterval.HMS
    tstruct: time.struct_time
    subsec: float


# hourly wake-ups for precise time tracking and early detection of DST changes
SET24 = frozenset(timeinterval.HMS([hour, 0, 0]) for hour in range(24))


class Cron(addons.AddonMainTask, block.SBlock):
    """
    Simple cron service.

    Do not use directly in circuits. It has a form of an SBlock only
    to allow debug messages and monitoring through the event interface.
    """

    def __init__(self, *args, utc, **kwargs):
        super().__init__(*args, **kwargs)
        self._timefunc = time.gmtime if utc else time.localtime
        self._alarms = {}   # {HMS: set_of_blocks}
        self._queue = None
        self._needs_reload = False

    def reload(self):
        """Reload the configuration after add_block/remove_block calls."""
        if self._needs_reload:
            if self._mtask is not None:
                self._queue.put_nowait(None)    # wake up the task, value does not matter
            self._needs_reload = False

    def add_block(self, hms, blk):
        """
        Add a block to be activated at given HMS.

        The block's 'recalc' method will be called at given time H:M:S
        and also when this service is started or reloaded.

        A TimeData object will be passed to the blk as its
        only argument.

        Don't forget to reload() after the last change.
        """
        if not hasattr(blk, 'recalc'):
            raise TypeError("{blk} is not compatible with the cron internal service")
        if not isinstance(hms, timeinterval.HMS):
            raise TypeError(f"argument 'hms': expected an HMS object, got {hms!r}")
        if hms in self._alarms:
            self._alarms[hms].add(blk)
        else:
            self._alarms[hms] = {blk}
            if hms not in SET24:
                self._needs_reload = True

    def remove_block(self, hms, blk):
        """
        Remove a blk for given HMS if it was registered.

        Do nothing if a registration was not found.

        A reload after the last change is recommended, but not
        strictly necessary.
        """
        if hms not in self._alarms:
            return
        self._alarms[hms].discard(blk)
        if not self._alarms[hms]:
            del self._alarms[hms]
            if hms not in SET24:
                self._needs_reload = True

    def get_current_time(self) -> TimeData:
        """Return the current time/date."""
        now = time.time()
        tstruct = self._timefunc(now)
        return TimeData(
            hms=timeinterval.HMS(tstruct),
            tstruct=tstruct,
            subsec=now % 1,     # don't want to import math just for the modf()
            )

    async def _maintask(self):
        """Recalculate registered blocks according to the schedule."""
        reset = False
        reload = True
        while True:
            # in outer loop:
            # - reset: DST begin/end or other computer clock related reason
            # - reload: self._alarms has changed
            now = self.get_current_time()
            if now.tstruct.tm_year < 2020:
                # this software did not exist back then
                raise RuntimeError("System clock is not set correctly.")
            if reset:
                for blk in set.union(*self._alarms.values()):    # all blocks
                    blk.recalc(now)
            if reload:
                timetable = sorted(set.union(set(self._alarms), SET24))
                tlen = len(timetable)
            next_idx = bisect.bisect_left(timetable, now.hms)
            reset = reload = False
            while True:
                # in inner loop: cycle through the timetable; break out to the outer loop
                # for a reset if time tracking is not accurate or for reload
                if next_idx == tlen:
                    next_idx = 0
                next_hms = timetable[next_idx]
                self.log("next wakeup at %s", next_hms)
                now = self.get_current_time()
                sleeptime = next_hms.seconds_from(now.hms) - now.subsec
                try:
                    await asyncio.wait_for(self._queue.get(), sleeptime)
                except asyncio.TimeoutError:
                    pass
                else:
                    reload = True
                    break
                now = self.get_current_time()
                if now.hms != next_hms:
                    # wrong time!
                    diff = now.hms.seconds_from(next_hms) + now.subsec
                    if diff > tconst.SEC_PER_DAY / 2:
                        diff -= tconst.SEC_PER_DAY
                    # diff > 0 = too late, diff < 0 = too early
                    reset = abs(diff) > _MAX_TRACKING_ERROR
                    self.warn(
                        "expected time: %s.000, current time: %s.%s, difference: %.3fs ",
                        next_hms, now.hms, format(now.subsec, '.3f')[2:], diff)
                    if reset:
                        self.warn("Resetting due to a time tracking error.")
                        break
                    if diff < 0.0:
                        # too early, everything should be fine after another sleep
                        continue
                if next_hms in self._alarms:
                    # recalc may alter the set we are iterating
                    for blk in list(self._alarms[next_hms]):
                        blk.recalc(now)
                next_idx += 1

    def init_regular(self):
        self.set_output(None)

    def start(self):
        super().start()
        self._queue = asyncio.Queue()

    def _event_get_schedule(self, **_data):
        """Return the internal scheduling data for debugging or monitoring."""
        return {
            str(hms): set(blk.name for blk in blkset)
            for hms, blkset in self._alarms.items()}



"""
Simple event-driven zero-delay logic/digital circuit simulation.

- - - - - -
Docs: https://edzed.readthedocs.io/en/latest/
Home: https://github.com/xitop/edzed/
"""

import asyncio
import fnmatch
import logging
import operator
from typing import Any, Iterator, Mapping, Optional

from . import addons
from . import block
from .blocklib import cblocks
from .blocklib import sblocks1
from .exceptions import EdzedError, EdzedInvalidState


__all__ = ['get_circuit', 'reset_circuit']

# limit for oscillation/instability detection:
_MAX_EVALS_PER_BLOCK = 3


_logger = logging.getLogger(__package__)
_current_circuit = None


def get_circuit() -> 'Circuit':
    """Get the current circuit. Create one if it does not exist."""
    global _current_circuit

    if _current_circuit is None:
        _current_circuit = Circuit()
    return _current_circuit


def reset_circuit() -> None:
    """
    Clear the circuit and create a new one.
    """
    global _current_circuit

    if _current_circuit is None:
        return
    try:
        _current_circuit.abort(EdzedError('forced circuit reset'))  # ignored if not running
    except Exception as err:
        # e.g. RuntimeError: Event loop is closed
        _logger.warning("reset_circuit(): %r error ignored", err)
    block.Event.instances.clear()
    _current_circuit = Circuit()


class Circuit:
    """
    A container of all blocks and their interconnections.

    Not to be instantiated directly, rather call edzed.get_circuit().

    There exist only one circuit at a time and all created blocks
    belongs to it.

    In this implementation, circuit blocks can be added, but not removed.
    """

    def __init__(self):
        self._blocks = {}           # all blocks belonging to this circuit by name
        self._simtask = None        # task running run_forever
        self._finalized = False     # no circuit modification after the initialization
        self._error = None          # exception that terminated the simulator or None
        self.persistent_dict = None # persistent state data back-end
        self.sblock_queue = None    # a Queue for notifying about changed SBlocks,
                                    # the queue will be created when simulation starts, because
                                    # it has a side effect of creating an event_loop if one
                                    # was not created yet. That may lead to errors in rare
                                    # scenarios with asyncio.new_event_loop().
        self._init_done = None      # an Event for wait_init() synchronization

    def is_current_task(self) -> bool:
        """Return True if the simulator task is the current asyncio task."""
        if self._simtask is None:
            return False
        try:
            return self._simtask == asyncio.current_task()
        except Exception:
            # for instance RuntimeError('no running event loop')
            return False

    def is_finalized(self) -> bool:
        """Return True only if finalize() was called."""
        return self._finalized

    def is_ready(self) -> bool:
        """Return True only if ready to accept external events."""
        return self._simtask is not None and self._error is None

    @property
    def error(self) -> Optional[BaseException]:
        return self._error

    async def _check_started(self) -> None:
        """Raise if the simulation was not started yet."""
        if self._simtask is None:
            # just created?
            await asyncio.sleep(0)
            if self._simtask is None:
                raise EdzedInvalidState("The simulation task was not started")

    async def wait_init(self) -> None:
        """Wait until a running circuit is fully initialized."""
        await self._check_started()
        await asyncio.wait(
            [asyncio.create_task(self._init_done.wait()), self._simtask],
            return_when=asyncio.FIRST_COMPLETED)
        if self._simtask.done():
            if self._simtask.cancelled():
                msg = "The simulation task is finished"
            else:
                # normal simtask exit is not possible
                msg = f"The simulation task failed with error: {self._simtask.exception()}"
            raise EdzedInvalidState(msg)

    def check_not_finalized(self) -> None:
        """Raise an error if the circuit has been finalized."""
        if self._error:
            # there is an even bigger problem
            raise EdzedInvalidState("The circuit was shut down")
        if self._finalized:
            raise EdzedInvalidState("No changes allowed in a finalized circuit")

    def set_persistent_data(self, persistent_dict: Optional[Mapping[str, Any]]) -> None:
        """Setup the persistent state data storage."""
        self.check_not_finalized()
        self.persistent_dict = persistent_dict

    def addblock(self, blk: block.Block) -> None:
        """
        Add a circuit block. Not an ad blocker :-)

        Application code does not call this method, because
        blocks register themselves automatically when created.
        """
        self.check_not_finalized()
        if not isinstance(blk, block.Block):
            raise TypeError(f"Expected a Block object, got {blk!r}")
        if blk.name in self._blocks:
            raise ValueError(f"Duplicate block name {blk.name}")
        self._blocks[blk.name] = blk

    def getblocks(self, btype: Optional[block.Block] = None) -> Iterator:
        """Return an iterator of all blocks or btype blocks only."""
        allblocks = self._blocks.values()
        if btype is None:
            return allblocks
        return (blk for blk in allblocks if isinstance(blk, btype))

    def findblock(self, name: str) -> block.Block:
        """Get block by name. Raise a KeyError when not found."""
        try:
            return self._blocks[name]
        except KeyError:
            # add an error message
            raise KeyError(f"Block '{name}' not found") from None

    def set_debug(self, value: bool, *args) -> int:
        """
        Set debug flag to given value (True/False) for selected blocks.
        """
        todo = set()
        for arg in args:
            if isinstance(arg, str):
                if any(ch in arg for ch in '*?['):
                    for blk in self.getblocks():
                        if fnmatch.fnmatchcase(blk.name, arg):
                            todo.add(blk)
                else:
                    todo.add(self.findblock(arg))
            elif isinstance(arg, block.Block):
                todo.add(arg)
            # issubclass works only for classes
            elif isinstance(arg, type) and issubclass(arg, block.Block):
                for blk in self.getblocks(arg):
                    todo.add(blk)
            else:
                raise TypeError(f"Expected block name, Block object or subclass, got {arg!r}")
        value = bool(value)
        for blk in todo:
            blk.debug = value
        return len(todo)

    def _check_persistent_data(self):
        """Check persistent state related settings."""
        persistent_blocks = [
            blk for blk in self.getblocks(addons.AddonPersistence) if blk.persistent]
        if self.persistent_dict is None:
            if persistent_blocks:
                _logger.warning(
                    "Disabling all persistent state, because the data storage was not set")
                for blk in persistent_blocks:
                    blk.persistent = False
        else:
            # clear the unused items
            used_keys = {blk.key for blk in persistent_blocks}
            for key in list(self.persistent_dict):
                if key not in used_keys:
                    _logger.info("Removing unused persistent state for '%s'", key)
                    del self.persistent_dict[key]

    def _validate_blk(self, blk):
        """
        Process a block specification. Return a valid block object.

        If the block is given by name, get the corresponding block.
        If a value is given, create a Const block from it.
        """
        if isinstance(blk, block.Const):
            return blk
        if isinstance(blk, str):
            # '_not_' + any 'NAME' except '_NAME'
            if blk.startswith('_') and blk not in self._blocks:
                # automatic blocks:
                if blk == '_ctrl':
                    return sblocks1.ControlBlock(
                        blk, desc="Simulation Control Block", _reserved=True)
                if blk.startswith('_not_') and blk[5:6] != '_':
                    return cblocks.Invert(
                        blk, desc=f"Inverted output of {blk}", _reserved=True
                        ).connect(blk[5:])
            return self.findblock(blk)
        if not isinstance(blk, block.Block):
            return block.Const(blk)
        if blk not in self.getblocks():
            raise ValueError(f"{blk} is not in the current circuit")
        return blk

    def _resolve_events(self):
        """Resolve temporary references by name in Event objects."""
        for event in block.Event.instances:
            if isinstance(event.dest, str):
                event.dest = self._validate_blk(event.dest)
            if not isinstance(event.dest, block.SBlock):
                raise ValueError(
                    f"Event destination block {event.dest} is not a sequential block")
        block.Event.instances.clear()   # free some memory

    def _finalize(self):
        """
        Complete the initialization of circuit block interconnections.

        For each block:
          - resolve temporary references by name
          - create Invert blocks for _not_name shortcuts
          - update connection data
        """
        def validate_output(iblk, oblk):
            """Validate oblk as a block to be connected to iblk."""
            try:
                return self._validate_blk(oblk)
            except Exception as err:
                fmt = f"Cannot connect {oblk} --> {iblk}: {{}}"
                err.args = (fmt.format(err.args[0]) if err.args else "<NO ARGS>", *err.args[1:])
                raise

        for btype in (block.CBlock, cblocks.Invert):
            # doing two passes because the first one may create new Invert blocks
            # some Invert blocks may be processed twice, that's acceptable
            for blk in list(self.getblocks(btype)):
                all_inputs = []
                for iname, inp in blk.inputs.items():
                    if isinstance(inp, tuple):
                        newinp = tuple(validate_output(blk, i) for i in inp)
                        all_inputs.extend(newinp)
                    else:
                        newinp = validate_output(blk, inp)
                        all_inputs.append(newinp)
                    blk.inputs[iname] = newinp

                for inp in all_inputs:
                    if not isinstance(inp, block.Const):
                        blk.iconnections.add(inp)
                        self._blocks[inp.name].oconnections.add(blk)

    def finalize(self):
        """A wrapper for _finalize()."""
        if not self._finalized:
            self._finalize()
            self._finalized = True

    async def _run_tasks(self, jobname, btt_list):
        """
        Run multiple tasks concurrently. Log errors.

        jobname is a description (e.g. "init" or "stop") used
        in messages. btt_list is a list of (block, task, timeout).
        It must not be empty.
        """
        assert btt_list
        errcnt = 0
        get_time = asyncio.get_running_loop().time
        start_time = get_time()
        for blk, task, timeout in sorted(btt_list, key=operator.itemgetter(2), reverse=True):
            # sorted from longest timeout
            if not task.done():
                try:
                    await asyncio.wait_for(task, timeout - get_time() + start_time)
                except asyncio.TimeoutError:
                    errcnt += 1
                    blk.warn("%s timeout, check timeout value (%.1fs)", jobname, timeout)
            if task.done() and not task.cancelled():
                err = task.exception()
                if err is not None:
                    errcnt += 1
                    blk.warn("%s error: %s", jobname, err, exc_info=err)
        if errcnt:
            _logger.error("%d block %s error(s) suppressed", errcnt, jobname)

    async def _init_sblocks_async(self):
        """Initialize all sequential blocks, the async part."""
        start_tasks = [
            (blk, asyncio.create_task(blk.init_async()), blk.init_timeout)
            for blk in self.getblocks(addons.AddonAsync)
            if blk.has_method('init_async') and blk.init_timeout > 0.0]
        if start_tasks:
            _logger.debug("Initializing async sequential blocks")
            await self._run_tasks("async init", start_tasks)

    @staticmethod
    def init_sblock(blk: block.Block) -> None:
        """
        Initialize a SBlock skipping any async code.

        Initialization order:
            - call init_from_persistent_data - if applicable
            - if not initialized, call init_regular
            - if not initialized, call init_from_value with the initdef
              value - if applicable

        After the init_sblock, a block:
            - MAY NOT be initialized, but
            - MUST be able to process events. An event is block's last
              chance to get its initialization. The simulation will
              fail unless all blocks are initialized properly.
        """
        if blk.is_initialized():
            return
        try:
            if isinstance(blk, addons.AddonPersistence) and blk.persistent:
                blk.init_from_persistent_data()
                if blk.is_initialized():
                    blk.log("initialized from saved state")
                    return
            blk.init_regular()
            if not blk.is_initialized() and  blk.has_method('init_from_value') \
                    and blk.initdef is not block.UNDEF:
                blk.init_from_value(blk.initdef)
        except Exception as err:
            # add the block name
            fmt = f"{blk}: error during initialization: {{}}"
            err.args = (fmt.format(err.args[0]) if err.args else "<NO ARGS>", *err.args[1:])
            raise

    def _init_sblocks_sync(self):
        """Initialize all sequential blocks, the non async part."""
        for blk in self.getblocks(block.SBlock):
            self.init_sblock(blk)
            if not blk.is_initialized():
                raise EdzedError(f"{blk}: not initialized")
        # save the internal states after initialization
        if self.persistent_dict is not None:
            for blk in self.getblocks(addons.AddonPersistence):
                blk.save_persistent_state()
        # clear the queue, the simulator knows that it needs to evaluate everything
        queue = self.sblock_queue
        while not queue.empty():
            queue.get_nowait()

    async def _stop_sblocks(self):
        """
        Stop all sequential blocks. Wait until all blocks are stopped.

        Suppress errors.
        """
        for blk in self.getblocks(block.SBlock):
            try:
                blk.stop()
            except Exception:
                _logger.error("%s: ignored error in stop()", blk, exc_info=True)
        await asyncio.sleep(0)
        wait_tasks = [
            (blk, asyncio.create_task(blk.stop_async()), blk.stop_timeout)
            for blk in self.getblocks(addons.AddonAsync)
            if blk.has_method('stop_async') and blk.stop_timeout > 0.0]
        if wait_tasks:
            _logger.debug("Waiting for async cleanup")
            await self._run_tasks("stop", wait_tasks)

    async def _simulate(self):
        """
        Simulate the circuit until cancelled.

        On first run evaluate all blocks with inputs (i.e. CBlocks).
        Then evaluate all blocks directly connected to:
            - changed sequential blocks
            - changed evaluated blocks
        """
        def select_blk(block_set):
            """
            Return one block from the given set with minimum number of
            its inputs connected to another block within the set.
            The set must not be empty.
            """
            min_idep = min_blk = None
            for blk in block_set:
                idep = sum(1 for inp in blk.iconnections if inp in block_set)
                if idep == 0:
                    # 0 is the absolute minimum, no need to search further
                    return blk
                if min_idep is None or idep < min_idep:
                    min_idep = idep
                    min_blk = blk
            return min_blk

        eval_limit = _MAX_EVALS_PER_BLOCK * len(self._blocks)
        eval_set = set(self.getblocks(block.CBlock))
        eval_cnt = 0
        queue = self.sblock_queue
        while True:
            if not eval_set and queue.empty():
                _logger.debug("%d block(s) evaluated, pausing", eval_cnt)
                blk = await queue.get()
                _logger.debug("output change in %s, resuming", blk)
                eval_cnt = 0
                eval_set |= blk.oconnections
            while not queue.empty():
                blk = queue.get_nowait()
                eval_set |= blk.oconnections
            if not eval_set:
                continue
            eval_cnt += 1
            if eval_cnt > eval_limit:
                raise EdzedError(
                    "Circuit instability detected (too many block evaluations)")
            if len(eval_set) == 1:
                blk = eval_set.pop()
            else:
                blk = select_blk(eval_set)
                eval_set.discard(blk)
            try:
                changed = blk.eval_block()
            except Exception as err:
                # add the block name
                fmt = f"{blk}: Error while evaluating block: {{}}"
                err.args = (fmt.format(err.args[0]) if err.args else "<NO ARGS>", *err.args[1:])
                raise
            if changed:
                eval_set |= blk.oconnections

    async def run_forever(self):    # no return value
        """
        Run the circuit simulation until the coroutine is cancelled.

        run_forever() never exits normally without an exception. It must
        be cancelled to stop the simulation.

        Please note that the cleanup could take some time - up to the
        max of SBlocks' stop_timeout values.

        When run_forever terminates, it cannot be invoked again.

        See also the ready attribute and wait_init().
        """
        if self._simtask is not None:
            if self._simtask.done():
                msg = "Cannot restart a finished simulation."
            else:
                msg = "The simulator is already running."
            raise EdzedInvalidState(msg)
        self._simtask = asyncio.current_task()
        blocks_started = False
        try:
            try:
                if self._error is not None:
                    raise self._error       # stop before start
                if not self._blocks:
                    raise EdzedError("The circuit is empty")

                _logger.debug("Initializing the circuit")
                self.sblock_queue = asyncio.Queue()
                self._init_done = asyncio.Event()
                self._check_persistent_data()
                self._resolve_events()
                self.finalize()

                _logger.debug("Setting up circuit blocks")
                blocks_started = True
                for blk in self.getblocks():
                    blk.start()
                # return control to the loop in order to run any tasks created by start()
                await asyncio.sleep(0)
                await self._init_sblocks_async()
                _logger.debug("Initializing sequential blocks")
                self._init_sblocks_sync()

                if self._error is None:
                    _logger.debug("Starting simulation")
                    self._init_done.set()
                    await self._simulate()
                    # not reached
                    assert False
            # CancelledError is derived from BaseException, (not Exception) in Python 3.8+
            except (Exception, asyncio.CancelledError) as err:
                if self._error is None:
                    self._error = err
            # Normally when a function from the try-except clause above calls abort(), the
            # abort() sets the self._error and cancels the task. The exception clause then
            # catches the cancellation.
            # But when a function calls abort() and also raises, the exception clause
            # catches the exception and the cancellation is left pending. For this
            # special edge case we must add a second except clause below.
            await asyncio.sleep(0)  # allow delivery of pending CancelledError if any
        except asyncio.CancelledError:
            pass

        if isinstance(self._error, asyncio.CancelledError):
            _logger.info("Normal circuit simulation stop")
        else:
            _logger.critical(
                "Fatal circuit simulation error: %s", self._error, exc_info=self._error)

        if blocks_started:
            # if blocks were started (at least some of them), they must be stopped
            # save the state first, because stop may invalidate the state information
            if self.persistent_dict is not None:
                for blk in self.getblocks(addons.AddonPersistence):
                    blk.save_persistent_state()
            await self._stop_sblocks()
        raise self._error

    def abort(self, exc: Exception) -> None:
        """
        Abort the circuit simulation due to an exception.

        abort() is necessary only when an ordinary exception wouldn't
        be propagated to the simulator. This is the case in asyncio
        tasks NOT using the SBlock._task_wrapper helper.

        The first error stops the simulation, that's why abort()
        delivers the exception only if the simulation hasn't received
        another exception already.

        abort() may be called even before the simulation begins.
        The start will then fail.
        """
        if self._error is not None:
            if not isinstance(exc, asyncio.CancelledError):
                _logger.warning("subsequent error: abort(%s)", exc)
            return
        if not isinstance(exc, BaseException):
            # one more reason to abort
            exc = TypeError(f'abort(): expected an exception, got {exc!r}')
        self._error = exc
        if self._simtask is not None and not self._simtask.done():
            self._simtask.cancel()

    async def shutdown(self) -> None:
        """
        Stop the simulation and wait until it finishes.
        """
        await self._check_started()
        if self.is_current_task():
            raise EdzedInvalidState("Cannot await the simulator task from the simulator task.")
        self.abort(asyncio.CancelledError('shutdown'))
        try:
            await self._simtask
        except asyncio.CancelledError:
            pass



"""
This module defines:
    - HMS class representing an idealized local clock time
    - MD class representing a date within a year
    - intervals using the HMS or MD objects as endpoints:
        - TimeInterval defines fixed time intervals within a day,
        - DateInterval defines fixed periods in a year.
      Intervals support the operation "value in interval".
"""

import collections.abc as cabc
import re
import time
from typing import Generator, Sequence, Union

from .tconst import *   # pylint: disable=wildcard-import


DAYS = (None, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)
# match group 1 = stripped match
RE_YEAR = re.compile(r'\s*(\d{4,})\s*', flags=re.ASCII)
RE_MONTH = re.compile(r'\s*([A-Z]{3,})\.?\s*', flags=re.ASCII|re.IGNORECASE)
RE_TIME = re.compile(r'\s*(\d{1,2}\s*:\s*\d{1,2}(\s*:\s*\d{1,2})?)\s*', flags=re.ASCII)
RE_DAY = re.compile(r'\s*(\d{1,2})\.?\s*', flags=re.ASCII)


def _cut(match):
    """
    Cut matched characters from the searched string.
    Join the remaining pieces with a space.
    """
    string, start, end = match.string, match.start(), match.end()
    if start == 0:
        return string[end:]
    if end == len(string):
        return string[:start]
    return ' '.join((string[:start], string[end:]))


# pylint: disable=invalid-name
def _ydt_convert(string, y=False, d=False, t=False):
    """
    Convert string to year (if y), month, day (if d), hour, minute, second (if t).

    The result is not fully validated.
    """

    year = month = day = None
    hms = (None, None, None)
    if t:
        match = RE_TIME.search(string)
        if not match:
            raise ValueError("missing time (H:M or H:M:S)")
        hms = [int(x) for x in match.group(1).split(':')]
        if len(hms) == 2:
            hms.append(0)
        string = _cut(match)
    if y:
        match = RE_YEAR.search(string)
        if not match:
            raise ValueError("missing year")
        year = int(match.group(1))
        string = _cut(match)
    if d:
        match = RE_MONTH.search(string)
        if not match:
            raise ValueError("missing month")
        name = match.group(1).capitalize()
        for i in range(1, 13):
            if MONTH_NAMES[i].startswith(name):
                month = i
                break
        else:
            raise ValueError(f"invalid month name: {match.group(1)!r}")
        string = _cut(match)
    if d:
        match = RE_DAY.search(string)
        if not match:
            raise ValueError("missing day of month")
        day = int(match.group(1))
        string = _cut(match)
    if string:
        raise ValueError(f"offending part: {string!r}")

    return (year, month, day, *hms)


def ydt_convert(string, *args, **kwargs):
    try:
        return _ydt_convert(string, *args, **kwargs)
    except Exception as err:
        raise ValueError(f"Could not convert {string!r}: {err}") from None


def _validate(y=None, md=None, hms=None):
    if y is not None:
        if y < 1970:
            # just a precaution, not a strict requirement
            raise ValueError(f"year {y} before start of the Unix Epoch (January 1, 1970)")
    if md is not None:
        month, day = md
        if not 1 <= month <= 12:
            raise ValueError(f"Invalid month number: {month}")
        if not 1 <= day <= DAYS[month]:
            raise ValueError(f"Invalid day in month number: {day}")
    if hms is not None:
        # leap second H:M:60 is accepted
        if not 0 <= hms[0] < 24 or not 0 <= hms[1] < 60 or not 0 <= hms[2] <= 60:
            raise ValueError(f"Time {hms} not between 0:0:0 and 23:59:59")


def _to_intlist(seq):
    try:
        return [int(x) for x in seq]
    except TypeError:
        raise TypeError(f"Cannot convert this type: '{seq!r}'")
    except ValueError:
        raise ValueError(f"Cannot convert this value: '{seq}'")


class HMS(tuple):
    """
    HMS stands for hour, minute, second. It is a 3-tuple representing
    a clock time within a day, always between (0,0,0) and (23,59,59).

    It can be directly compared with other tuples of the same format.

    Individual fields are accessible as hour, minute and second
    attributes.
    """

    def __new__(
            cls,
            hms: Union[None, 'HMS', time.struct_time, int, str, Sequence[int]] = None):
        """
        HMS() or HMS(None) = use the current local time of day
        HMS(hms) = return the existing instance
        HMS(time.struct_time) = use the provided time
        HMS(integer) = convert the seconds counted from midnight 00:00:00
        HMS(string) = convert from "HH:MM" or "HH:MM:SS" format
        HMS(iterable with 2 elements) = use as hour and minute value
        HMS(iterable with 3 elements) = use as hour, minute, and second
                                        (avoid unordered iterables)
        """
        if isinstance(hms, cls):
            # OK, it is immutable
            return hms
        if hms is None:
            hms = time.localtime()
        if isinstance(hms, time.struct_time):
            hms3 = (hms.tm_hour, hms.tm_min, hms.tm_sec)
        elif isinstance(hms, int):
            seconds = hms % SEC_PER_DAY
            hours, seconds = divmod(seconds, SEC_PER_HOUR)
            minutes, seconds = divmod(seconds, SEC_PER_MIN)
            hms3 = (hours, minutes, seconds)
        else:
            if isinstance(hms, str):
                hms3 = ydt_convert(hms, t=True)[3:6]
            else:
                hms3 = _to_intlist(hms)
                if len(hms3) == 2:
                    hms3.append(0)
                elif len(hms3) != 3:
                    raise ValueError(f"Invalid time specification: '{hms}'")
            _validate(hms=hms3)
        return super().__new__(cls, hms3)

    @property
    def hour(self) -> int:
        return self[0]

    @property
    def minute(self) -> int:
        return self[1]

    @property
    def second(self) -> int:
        return self[2]

    def seconds(self) -> int:
        """Return seconds counted from 00:00:00."""
        return self[0]*SEC_PER_HOUR + self[1]*SEC_PER_MIN + self[2]

    def seconds_from(self: 'HMS', other: 'HMS') -> int:
        """
        Return seconds from other HMS to this HMS.

        The result is always >= 0:
            HMS('10:30').seconds_from(HMS('9:30')) ==  1 * SEC_PER_HOUR
        because 10:30 is one hour after 9:30, but:
            HMS('9:30').seconds_from(HMS('10:30')) == 23 * SEC_PER_HOUR
        because 9:30 is 23 hours after 10:30.
        """
        sec = self.seconds() - other.seconds()
        return sec + SEC_PER_DAY if sec < 0 else sec

    def __str__(self):
        return f"{self.hour:02d}:{self.minute:02d}:{self.second:02d}"

    def __repr__(self):
        cls = type(self)
        return f"{cls.__module__}.{cls.__qualname__}('{self}')"


class MD(tuple):
    """
    MD stands for month and day. It is a 2-tuple representing a calendar
    date, i.e. always between (1,1) and (12,31).

    It can be directly compared with other tuples of the same format.

    Individual fields are accessible as day and month attributes.
    """

    def __new__(cls, md: Union[None, 'MD', time.struct_time, str, Sequence[int]] = None):
        """
        MD() or MD(None) = use the current local date
        MD(md) = return the existing instance
        MD(time.struct_time) = use the provided date
        MD(string) = convert from a string
        MD(iterable with 2 elements) = use as day and month numeric values
                                       (avoid unordered iterables)

        When building from a string, the month must be written as a three
        letter English acronym in order to avoid ambiguities. The rules:
            - day = 1 or 2 digits
            - month = 3 letters in lower, upper or mixed case
            - there may be one period (full stop) appended
              directly to the day or the month
            - the date consists of day and month in this or in
              reversed order
            - leading, trailing whitespace and any whitespace between
              the day and the month is ignored
        Examples:
            09.Jun   9.jun.   9JUN   09 Jun   Jun9   JUN 09   Jun.9.
        """
        if isinstance(md, cls):
            # OK, it is immutable
            return md
        if md is None:
            md = time.localtime()
        if isinstance(md, time.struct_time):
            md2 = (md.tm_mon, md.tm_mday)
        else:
            if isinstance(md, str):
                md2 = ydt_convert(md, d=True)[1:3]
            else:
                md2 = _to_intlist(md)
                if len(md) != 2:
                    raise ValueError(f"Invalid month, day specification: '{md}'")
            _validate(md=md2)
        return super().__new__(cls, md2)

    @property
    def month(self) -> int:
        return self[0]

    @property
    def day(self) -> int:
        return self[1]

    def __str__(self):
        """Output format is 'Jun.09'"""
        return f'{MONTH_NAMES[self.month][:3]}.{self.day:02d}'

    def __repr__(self):
        cls = type(self)
        return f"{cls.__module__}.{cls.__qualname__}('{self}')"


class YDT(tuple):
    """
    YDT stands for year, date (as in MD) and time (as in HMS).
    It is a 6-tuple of integers: (year, month, date, hour, minute, second).

    It can be directly compared with other tuples of the same format.

    Individual fields are accessible as attributes.
    """

    def __new__(cls, ydt: Union[None, 'YDT', time.struct_time, str, Sequence[int]] = None):
        """
        YDT() or YDT(None) = use the current local date and time
        YDT(ydt) = return the existing instance
        YDT(time.struct_time) = use the provided time
        YDT(string) = convert from a string
        YDT(iterable with 5 elements) = use as year, month, day, hour and minute value
        YDT(iterable with 6 elements) = use as year, ... second
                                        (avoid unordered iterables)
        """
        if isinstance(ydt, cls):
            # OK, it is immutable
            return ydt
        if ydt is None:
            ydt = time.localtime()
        if isinstance(ydt, time.struct_time):
            ydt6 = tuple(ydt[0:6])
        else:
            if isinstance(ydt, str):
                ydt6 = ydt_convert(ydt, y=True, d=True, t=True)
            else:
                ydt6 = _to_intlist(ydt)
                if len(ydt6) == 5:
                    ydt6.append(0)
                elif len(ydt6) != 6:
                    raise ValueError(f"Invalid date and time specification: '{ydt}'")
                _validate(y=ydt6[0], md=ydt6[1:3], hms=ydt6[3:6])
        return super().__new__(cls, ydt6)

    @property
    def year(self) -> int:
        return self[0]

    @property
    def month(self) -> int:
        return self[1]

    @property
    def day(self) -> int:
        return self[2]

    @property
    def hour(self) -> int:
        return self[3]

    @property
    def minute(self) -> int:
        return self[4]

    @property
    def second(self) -> int:
        return self[5]

    def __str__(self):
        return f"{self.year} {MONTH_NAMES[self.month][:3]}.{self.day:02d} " \
               f"{self.hour:02d}:{self.minute:02d}:{self.second:02d}"

    def __repr__(self):
        cls = type(self)
        return f"{cls.__module__}.{cls.__qualname__}('{self}')"


SEPCHAR = ','
RANGECHAR = '-'
class _Interval:
    """
    The common part of TimeInterval and DateInterval.

    Warning: time/date intervals do not follow strict mathematic
    interval definition.
    """

    # https://en.wikipedia.org/wiki/Interval_(mathematics)#Terminology
    # the subintervals are always left-closed
    _RCLOSED_INTERVAL = False # are the subintervals also right-closed?

    @staticmethod
    def _convert(val: Union[str, Sequence]):
        """Convert interval endpoint."""

    def __init__(self, ivalue: Union[None, '_Interval', str, Sequence] = None):
        """
        _Interval() = empty interval
        _Interval(interval) = copy of interval
        _Interval('string') = converted from a string containing comma
            separated ranges:
                "FROM1-TO1,FROM2-TO2".
            If _RCLOSED_INTERVAL is True a single value is accepted
            as well, e.g.:
                "FROM1-TO1,VALUE2"
            is equivalent to "FROM1-TO1, VALUE2-VALUE2".
            The input string may contain whitespace around any value.
        _Interval([[from1, to1], [from2, to2], ...]) = create from
            a sequence of pairs
        """
        if ivalue is None:
            self._interval = []
            return
        if isinstance(ivalue, type(self)):
            # pylint: disable=protected-access
            self._interval = ivalue._interval.copy()
            return
        if isinstance(ivalue, str):
            self._interval = []
            ivalue = ivalue.strip()
            if not ivalue:
                return
            for rstr in ivalue.split(SEPCHAR):
                try:
                    split_here = rstr.index(RANGECHAR, 1, -1)
                except ValueError:
                    if not self._RCLOSED_INTERVAL:
                        raise ValueError(f"Invalid range '{rstr}'") from None
                    low = high = self._convert(rstr)
                else:
                    low = self._convert(rstr[:split_here])
                    high = self._convert(rstr[split_here+1:])
                self._interval.append((low, high))
            return
        if isinstance(ivalue, (cabc.Sequence, cabc.Iterator)):
            self._interval = [(self._convert(low), self._convert(high)) for low, high in ivalue]
            return
        raise TypeError("Invalid interval value")

    def range_endpoints(self) -> Generator:
        """Yield all range start and stop values."""
        for low, high in self._interval:
            yield low
            yield high

    @staticmethod
    def _cmp_open(low, item, high):
        """
        The ranges are left-closed and right-open intervals, i.e.
        value is in interval if and only if start <= value < stop
        """
        if low < high:
            return low <= item < high
        return low <= item or item < high   # low <= item < MAX or MIN <= item < high

    @staticmethod
    def _cmp_closed(low, item, high):
        """
        The ranges are closed intervals, i.e.
        value is in interval if and only if start <= value <= stop
        """
        if low <= high:
            return low <= item <= high
        return low <= item or item <= high

    def _cmp(self, *args):
        return (self._cmp_closed if self._RCLOSED_INTERVAL else self._cmp_open)(*args)

    def __contains__(self, item):
        return any(self._cmp(low, item, high) for low, high in self._interval)

    def as_list(self):
        """
        Return the intervals as a nested list of integers.

        The output is suitable as an input argument.
        """
        return [[list(low), list(high)] for low, high in self._interval]

    def __str__(self):
        def fmt(low, high):
            if self._RCLOSED_INTERVAL and low == high:
                return str(low)
            return f"{low}{RANGECHAR}{high}"
        return f'{SEPCHAR} '.join(fmt(low, high) for low, high in self._interval)

    def __repr__(self):
        cls = type(self)
        return f"{cls.__module__}.{cls.__qualname__}('{self}')"


class TimeInterval(_Interval):
    """
    List of time ranges.

    See the HMS class for time formats.

    The whole day is 00:00 - 00:00.
    """

    _RCLOSED_INTERVAL = False
    def _convert(self, val):
        if isinstance(val, int):
            # HMS can convert int, but in this context it is almost
            # certainly a result of a malformed input. Compare:
            #   incorrect: [[h1,m1],[h2,m2]]
            #   correct:  [[[h1,m1],[h2,m2]]]
            raise TypeError(
                f"Expected was [H, M] or [H, M, S], but got an int ({val}); "
                "check the structure of the input argument")
        return HMS(val)


class DateInterval(_Interval):
    """
    List of date ranges and single dates.

    See the MD class for date formats.
    """

    _RCLOSED_INTERVAL = True
    _convert = MD


class DateTimeInterval(_Interval):
    """
    List of date+time ranges.

    See the YDT class.
    """

    _RCLOSED_INTERVAL = False
    _convert = YDT

    @staticmethod
    def _cmp_open(low, item, high):
        """Compare function for non-repeating intervals."""
        return low <= item < high



"""
Add-ons extending SBlocks' capabilities.

For each add-on there is usually also a supporting code in
the simulator.

- - - - - -
Docs: https://edzed.readthedocs.io/en/latest/
Home: https://github.com/xitop/edzed/
"""

import abc
import asyncio
from typing import Any, Awaitable, Mapping, Union

from . import block
from .exceptions import EdzedError


__all__ = ['AddonAsync', 'AddonMainTask', 'AddonPersistence']


class AddonPersistence(block.Addon, metaclass=abc.ABCMeta):
    """
    Add support for persistent state to a SBlock.
    """

    def __init__(self, *args, persistent: bool = False, sync_state: bool = True, **kwargs):
        self.persistent = bool(persistent)
        self.sync_state = bool(sync_state)
        super().__init__(*args, **kwargs)
        # str(self) (as defined in superclass!) is used as a key instead of
        # just the name, because it contains also the block type name.
        self.key = str(self)

    # TODO in Python3.8+ (see Block.event for an explanation):
    #   def event(self, etype: Union[str, block.EventType], /, **data) -> Any:
    # pylint: disable=no-method-argument
    def event(*args, **data) -> Any:
        """Save persistent state after a possible state change."""
        self, etype = args
        try:
            # TODO in Python3.8+: super().event
            retval = super(AddonPersistence, self).event(etype, **data)
        except Exception:
            if self.persistent and not self.circuit.is_ready():
                # The internal state data may be corrupted, because it looks like
                # event() decided to stop the simulation in reaction to this exception.
                # (Never mind if it wasn't this exception but some previous one.)
                self.warn("Disabling persistent state due to an error")
                self.persistent = False
            raise
        if self.persistent and self.sync_state:
            self.save_persistent_state()
        return retval

    def save_persistent_state(self) -> None:
        """
        Save the state to persistent storage if enabled. Suppress errors.
        """
        if not self.persistent:
            return
        try:
            self.circuit.persistent_dict[self.key] = self.get_state()
        except Exception as err:
            self.warn("Persistent data save error: %s", err)
            self.circuit.persistent_dict.pop(self.key, None)  # remove stale data

    @abc.abstractmethod
    def _restore_state(self, state: Any) -> None:
        """
        Initialize by restoring the state (low-level).
        """

    def init_from_persistent_data(self) -> None:
        """
        Initialize by restoring the saved state (high-level).

        Load the state from persistent storage and apply it.

        IMPORTANT: output change events are temporarily disabled
        when loading the saved state.

        Errors are suppressed.
        """
        try:
            state = self.circuit.persistent_dict[self.key]
        except KeyError:
            return
        except Exception as err:
            self.warn("Persistent data retrieval error: %s", err)
            return
        try:
            self._restore_state(state)
        except Exception as err:
            self.warn("Error restoring saved state: %s; state: %s", err, state)

    def get_conf(self) -> Mapping[str, Any]:
        return {
            'persistent': self.persistent,
            **super().get_conf()
        }


DEFAULT_INIT_TIMEOUT = 10.0
DEFAULT_STOP_TIMEOUT = 10.0
class AddonAsync(block.Addon):
    """
    Asynchronous support add-on.
    """

    def __init__(self, *args, **kwargs):
        """
        Parameters:
            init_timeout:
                init_async() timeout in seconds (default=10),
                valid only if the init_async method is defined.
                Value 0.0 (or negative) disables the init_async().
            stop_timeout:
                stop_async() timeout in seconds (default=10),
                valid only if the stop_async() method is defined.
                Value 0.0 (or negative) disables the stop_async().

        All timeouts are in seconds (int or float). Value None or a
        missing argument are replaced by the default timeout.
        """
        init = self.has_method('init_async')
        if init:
            self.init_timeout = kwargs.pop('init_timeout', None)
        stop = self.has_method('stop_async')
        if stop:
            self.stop_timeout = kwargs.pop('stop_timeout', None)
        super().__init__(*args, **kwargs)
        # cannot use self.log before Block.__init__()
        if init and self.init_timeout is None:
            self.log("init_timeout not set, default is %.3fs", DEFAULT_INIT_TIMEOUT)
            self.init_timeout = DEFAULT_INIT_TIMEOUT
        if stop and self.stop_timeout is None:
            self.log("stop_timeout not set, default is %.3fs", DEFAULT_STOP_TIMEOUT)
            self.stop_timeout = DEFAULT_STOP_TIMEOUT

    async def _task_wrapper(self, coro: Awaitable, is_service: bool = False) -> Any:
        """
        A coroutine wrapper delivering exceptions to the simulator.

        Couroutines marked as services (is_service=True) are supposed
        to run until cancelled - a normal exit is treated as an error.

        Cancellation is not considered an error.
        """
        try:
            retval = await coro
            if is_service:
                raise EdzedError("coroutine providing a service has exited")
        # not needed in Python 3.8+
        except asyncio.CancelledError:  # pylint: disable=try-except-raise
            raise
        except Exception as err:
            # add context to the error message
            fmt = f"{self}: error in {coro.__qualname__}: {{}}"
            err.args = (fmt.format(err.args[0] if err.args else "<NO ARGS>"), *err.args[1:])
            self.circuit.abort(err)
            raise
        return retval


class AddonMainTask(AddonAsync, metaclass=abc.ABCMeta):
    """
    An add-on running a '_maintask' from start() till stop().
    """

    def __init__(self, *args, **kwargs):
        self._mtask = None
        super().__init__(*args, **kwargs)

    @abc.abstractmethod
    async def _maintask(self):
        pass

    def start(self) -> None:
        super().start()
        assert self._mtask is None
        self._mtask = asyncio.create_task(self._task_wrapper(self._maintask(), is_service=True))

    async def stop_async(self) -> None:
        if self._mtask is None:
            return
        self._mtask.cancel()
        try:
            await self._mtask
        except asyncio.CancelledError:
            pass
        finally:
            self._mtask = None
        await super().stop_async()



"""
Periodic events at fixed time/date.

- - - - - -
Docs: https://edzed.readthedocs.io/en/latest/
Home: https://github.com/xitop/edzed/
"""

from .. import addons
from .. import block
from .. import cron
from .. import simulator
from ..utils import timeinterval


__all__ = ['TimeDate', 'TimeSpan']


def _get_cron(utc):
    circuit = simulator.get_circuit()
    name = '_cron_utc' if utc else '_cron_local'
    try:
        return circuit.findblock(name)
    except KeyError:
        return cron.Cron(name, utc=utc, _reserved=True)


class TimeDate(addons.AddonPersistence, block.SBlock):
    """
    Block for periodic events at given time/date.
    """

    def __init__(self, *args, times=None, dates=None, weekdays=None, utc=False, **kwargs):
        self._cron = _get_cron(bool(utc))
        self._times = self._dates = self._weekdays = None
        # we build the initdef from times, dates, weekdays in order to simplify the usage
        if 'initdef' in kwargs:
            raise TypeError(
                f"'initdef' is an invalid keyword argument for {type(self).__name__}")
        initdef = self.parse(times, dates, weekdays)
        super().__init__(*args, initdef=initdef, **kwargs)

    @classmethod
    def parse(cls, times, dates, weekdays) -> dict:
        """Return the values in a normalized form."""
        return cls._export3(*cls._parse3(times, dates, weekdays))

    @staticmethod
    def _parse3(times, dates, weekdays):
        if times is not None:
            times = timeinterval.TimeInterval(times)
        if dates is not None:
            dates = timeinterval.DateInterval(dates)
        if weekdays is not None:
            if isinstance(weekdays, str):
                weekdays = [int(x) for x in weekdays if x != ' ']
            if not all(0 <= x <= 7 for x in weekdays):
                raise ValueError(
                    "Only numbers 0 or 7 (Sun), 1 (Mon), ... 6(Sat) are accepted as weekdays")
            weekdays = frozenset(0 if x == 7 else x for x in weekdays)
        return times, dates, weekdays

    @staticmethod
    def _export3(times, dates, weekdays):
        return {
            'times': None if times is None else times.as_list(),
            'dates': None if dates is None else dates.as_list(),
            'weekdays': None if weekdays is None else sorted(weekdays),
        }

    def get_state(self):
        return self._export3(self._times, self._dates, self._weekdays)

    def recalc(self, now: cron.TimeData):
        """Update the output."""
        tnone, dnone, wnone = self._times is None, self._dates is None, self._weekdays is None
        self.set_output(
            not (tnone and dnone and wnone)
            and (tnone or now.hms in self._times)
            and (dnone or timeinterval.MD(now.tstruct) in self._dates)
            # convert 0-6 (Mon-Sun) to 0-6 (Sun-Sat)
            and (wnone or (now.tstruct.tm_wday + 1) % 7 in self._weekdays))

    def _event_reconfig(self, *, times=None, dates=None, weekdays=None, **_data):
        """Reconfigure the block."""
        if self._times is not None:
            for hms in self._times.range_endpoints():
                self._cron.remove_block(hms, self)
        self._times, self._dates, self._weekdays = self._parse3(times, dates, weekdays)
        if self._times is not None:
            for hms in self._times.range_endpoints():
                self._cron.add_block(hms, self)
        # sometimes it is necessary, sometimes not, but it is easier
        # to always add 0:0:0 than to analyze the arguments
        self._cron.add_block(timeinterval.HMS([0, 0, 0]), self)
        self._cron.reload()
        self.recalc(self._cron.get_current_time())

    def init_from_value(self, value):
        self._event_reconfig(**value)

    _restore_state = init_from_value


class TimeSpan(addons.AddonPersistence, block.SBlock):
    """
    Block active between start and stop time/date.
    """

    def __init__(self, *args, span=(), utc=False, **kwargs):
        self._cron = _get_cron(bool(utc))
        self._span = timeinterval.DateTimeInterval(())
        # we build the initdef
        if 'initdef' in kwargs:
            raise TypeError(
                f"'initdef' is an invalid keyword argument for {type(self).__name__}")
        initdef = self.parse(span)
        super().__init__(*args, initdef=initdef, **kwargs)

    @classmethod
    def parse(cls, span) -> list:
        """Return the value in a normalized form."""
        return timeinterval.DateTimeInterval(span).as_list()

    def get_state(self):
        return self._span.as_list()

    def recalc(self, now: cron.TimeData):
        """Update the output."""
        self.set_output(timeinterval.YDT(now.tstruct) in self._span)

    def _event_reconfig(self, *, span=(), **_data):
        """Reconfigure the block."""
        for ydt in self._span.range_endpoints():
            self._cron.remove_block(timeinterval.HMS(ydt[3:6]), self)
        self._span = timeinterval.DateTimeInterval(span)
        now = self._cron.get_current_time()
        now_ymd = tuple(now.tstruct[0:3])
        for ydt in self._span.range_endpoints():
            if ydt[0:3] >= now_ymd:
                # future events only
                self._cron.add_block(timeinterval.HMS(ydt[3:6]), self)
        self._cron.reload()
        self.recalc(now)

    def init_from_value(self, value):
        self._event_reconfig(span=value)

    _restore_state = init_from_value



"""
Sequential blocks for general use.

- - - - - -
Docs: https://edzed.readthedocs.io/en/latest/
Home: https://github.com/xitop/edzed/
"""

import asyncio

from .. import addons
from .. import block
from ..exceptions import EdzedError
from .. import fsm


__all__ = ['ControlBlock', 'Counter', 'Repeat', 'ValuePoll']


class ControlBlock(block.SBlock):
    """
    Simulator control block.
    """

    def _event_shutdown(self, *, source='<no-source-data>', **_data):
        exc = asyncio.CancelledError(f"{self}: shutdown requested by '{source}'")
        self.circuit.abort(exc)

    def _event_error(self, *, source='<no-source-data>', error='<no-error-data>', **_data):
        exc = EdzedError(f"{self}: error reported by '{source}': {error!r}")
        if isinstance(error, Exception):
            exc.__cause__ = error
        self.circuit.abort(exc)

    def init_regular(self):
        self.set_output(None)


class Counter(addons.AddonPersistence, block.SBlock):
    """
    Counter. If modulo is set to a number M, count modulo M.
    """

    def __init__(self, *args, modulo=None, initdef=0, **kwargs):
        """
        Set the optional modulo and the initial counter value (default = 0).
        """
        if modulo == 0:
            raise ValueError("modulo must not be zero")
        self._mod = modulo
        super().__init__(*args, initdef=initdef, **kwargs)

    def _setmod(self, value):
        self.set_output(value if self._mod is None else value % self._mod)

    def _event_inc(self, *, amount=1, **_data):
        self._setmod(self._output + amount)

    def _event_dec(self, *, amount=1, **_data):
        self._setmod(self._output - amount)

    def _event_put(self, *, value, **_data):
        self._setmod(value)

    init_from_value = _setmod
    _restore_state = _setmod


class Repeat(addons.AddonMainTask, block.SBlock):
    """
    Periodically repeat the last received event.
    """

    def __init__(self, *args, dest, etype='put', interval, **kwargs):
        self._repeated_event = block.Event(dest, etype)
        interval = fsm.convert_duration(interval)
        if interval is None or interval <= 0.0:
            raise ValueError("interval must be positive")
        self._interval = interval
        self._queue = None
        super().__init__(*args, **kwargs)

    async def _maintask(self):
        repeat = False
        self.set_output(False)
        data = await self._queue.get()
        data['orig-source'] = data.get('source')
        while True:
            self.set_output(repeat)
            self._repeated_event.send(self, **data, repeat=repeat)
            repeat = True
            try:
                data = await asyncio.wait_for(self._queue.get(), self._interval)
            except asyncio.TimeoutError:
                pass
            else:
                repeat = False

    def _event(self, etype, data):
        if etype == self._repeated_event.etype:
            self._queue.put_nowait(data)

    def start(self):
        super().start()
        self._queue = asyncio.Queue()


class ValuePoll(addons.AddonMainTask, block.SBlock):
    """
    A source of measured or computed values.
    """

    def __init__(self, *args, func, interval, **kwargs):
        self._func = func
        self._interval = fsm.convert_duration(interval)
        self._init_done = asyncio.Event()
        super().__init__(*args, **kwargs)

    async def _maintask(self):
        """Data acquisition task: repeatedly obtain a value."""
        initialized = False
        while True:
            value = self._func()
            if asyncio.iscoroutine(value):
                value = await value
            if value is not block.UNDEF:
                self.set_output(value)
                if not initialized:
                    self._init_done.set()
                    initialized = True
            await asyncio.sleep(self._interval)

    async def init_async(self):
        await self._init_done.wait()

    def init_from_value(self, value):
        self.set_output(value)



"""
A library for building small automated systems.

The edzed package contains:
 - classes for creating combinational and sequential blocks
 - methods for building a circuit by connecting the blocks
 - a simple event-driven zero-delay digital circuit simulator

- - - - - -
Copyright (c) 2019-2020 Vlado Potisk <edzed@poti.sk>.

Released under the MIT License.

Docs: https://edzed.readthedocs.io/en/latest/
Home: https://github.com/xitop/edzed/
"""

# pylint: disable=wildcard-import,undefined-variable
from .exceptions import *
from .block import *
from .addons import *
from .simulator import *
from .fsm import *
# .demo is not imported to edzed
from .blocklib.cblocks import *
from .blocklib.filters import *
from .blocklib.fsms import *
from .blocklib.sblocks1 import *
from .blocklib.sblocks2 import *
from .blocklib.timedate import *

__all__ = (
    addons.__all__
    + block.__all__
    + exceptions.__all__
    + fsm.__all__
    + simulator.__all__
    + blocklib.cblocks.__all__
    + blocklib.filters.__all__
    + blocklib.fsms.__all__
    + blocklib.sblocks1.__all__
    + blocklib.sblocks2.__all__
    + blocklib.timedate.__all__
    )



"""
Convert timestamps between event loop and system.
"""

import asyncio
import time

def _get_timediff():
    """Return the difference between loop and Unix time bases."""
    loopnow = asyncio.get_running_loop().time()
    unixnow = time.time()
    return unixnow - loopnow

def loop_to_unixtime(looptime, timediff=None):
    """Convert event loop time to standard Unix time."""
    if timediff is None:
        timediff = _get_timediff()
    return looptime + timediff

def unix_to_looptime(unixtime, timediff=None):
    """Convert standard Unix time to event loop time."""
    if timediff is None:
        timediff = _get_timediff()
    return unixtime - timediff



"""
A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='edzed',
    version='20.9.10',
    description='A library for building small automated systems',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/xitop/edzed',
    author='Vlado Potisk',
    author_email='edzed@poti.sk',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    project_urls={
        'Source': 'https://github.com/xitop/edzed',
        'Documentation': 'https://edzed.readthedocs.io/en/latest/',
    },

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'test*']),

    python_requires='>=3.7',

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[],
)



"""
Conversion routines for time spans using multiple units.

Example: "20h15m10" = 20 hours + 15 minutes + 10 seconds = 72910 seconds
"""

import re
from typing import Union

from .tconst import *   # pylint: disable=wildcard-import, unused-wildcard-import

# pylint: disable=invalid-name
def timestr(seconds: Union[int, float]) -> str:
    """
    Return seconds as a string using d, h, m and s units.

    Minutes and seconds are always present in the result.
    Days and hours are prepended only when needed.

    This is an inverse function to convert() below.
    Partial seconds are formatted to 3 decimal places.
    """
    if seconds < 0:
        raise ValueError("Number of seconds cannot be negative")
    d, s = divmod(seconds, SEC_PER_DAY)
    h, s = divmod(s, SEC_PER_HOUR)
    m, s = divmod(s, SEC_PER_MIN)
    parts = []
    if d:
        parts.append(f"{int(d)}d")
    if d or h:
        parts.append(f"{int(h)}h")
    parts.append(f"{int(m)}m")
    # seconds = original value; s = 0 to 60 seconds
    parts.append(f"{s:.3f}s" if isinstance(seconds, float) else f"{s}s")
    return ''.join(parts)


_RE_TIME = r'(?:(\d+)\s*d)?\s*(?:(\d+)\s*h)?\s*(?:(\d+)\s*m)?\s*(?:([\d.]+)\s*s?)?'
_REGEXP_TIME = re.compile(_RE_TIME, re.ASCII | re.IGNORECASE)

def convert(tstr: str) -> float:
    """
    Convert string to number of seconds. Return float.

    Format:
        TIMESTR = [DAYS] [HOURS] [MINUTES] [SECONDS]
    where:
        DAYS = <int> "D"
        HOURS = <int> "H"
        MINUTES = <int> "M"
        SECONDS =  <int or float> ["S"]
    Notes:
        - whitespace around numbers and units is allowed
        - numbers do not have to be normalized, e.g. "72h" is OK
        - unit symbols D, H, M, S may be entered in upper or lower case
        - negative values are not allowed
        - float values with exponents are not supported
    """
    match = _REGEXP_TIME.fullmatch(tstr.strip())
    if match is None:
        raise ValueError("Invalid time representation")
    d, h, m, s = match.groups()
    result = 0.0
    if d:
        result += int(d) * SEC_PER_DAY
    if h:
        result += int(h) * SEC_PER_HOUR
    if m:
        result += int(m) * SEC_PER_MIN
    if s:
        result += float(s)  # may raise ValueError, the regexp is not strict enough
    return result



"""
A simple interactive utility for learning and evaluating edzed.

WARNING: Do not use for anything else. Contains "eval <user input>".

Usage:
    1. append following lines to a python file with an edzed
    circuit definition (e.g. circuit1.py):
    --- begin ---
    if __name__ == '__main__':
        import asyncio  # if not imported at the top
        import edzed.demo
        asyncio.run(edzed.demo.run_demo())
    --- end ---
    2. run the demo:
        python3 circuit1.py

- - - - - -
Docs: https://edzed.readthedocs.io/en/latest/
Home: https://github.com/xitop/edzed/
"""

import ast
import asyncio
from dataclasses import dataclass
import logging
import sys
from typing import Callable

from . simulator import get_circuit

if sys.platform == 'win32':
    # Python 3.7 only, ProactorEventLoop is the default on 3.8+
    asyncio.set_event_loop(asyncio.ProactorEventLoop())


__all__ = ['run_demo']

LIMIT = 4096

HELP = """\
Control commands:
    h[elp] or ?                 -- show this help
    exit
    eval <python_expression>
Circuit evaluation commands:
    c[debug] 1|0                -- circuit simulator's debug messages on|off
    d[ebug] <blockname> 1|0     -- block's debug messages on|off
    e[vent] <blockname> <type> [{'name':value, ...}]
                                -- send event
    i[nfo] <blockname>          -- print block's properties
    l[ist]                      -- list all blocks
    p[ut] <blockname> <value>   -- send 'put' event
    s[how] <blockname>          -- print current state and output
Command history:
    !!                          -- repeat last command
    !0 to !9                    -- repeat command N
    !?                          -- print history
"""

def _cmd_cdebug(value):
    if value == '0':
        logging.getLogger().setLevel(logging.INFO)
        print("circuit simulator debug off")
        return
    if value == '1':
        logging.getLogger().setLevel(logging.DEBUG)
        print("circuit simulator debug on")
        return
    raise ValueError("Argument must be 0 (debug off) or 1 (debug on)")


def _cmd_debug(blk, value):
    if value == '0':
        blk.debug = False
        print(f"{blk}: debug off")
        return
    if value == '1':
        print(f"{blk}: debug on")
        blk.debug = True
        return
    raise ValueError("Argument must be 0 (debug off) or 1 (debug on)")


def _cmd_eval(expr):
    result = eval(expr)         # pylint: disable=eval-used
    print(f"result: {result}")


def _cmd_event(blk, etype, data=None):
    data = {} if data is None else ast.literal_eval(data)
    retval = blk.event(etype, **data)
    print(f"event() returned: {retval}")
    _cmd_show(blk)


def _cmd_help():
    print(HELP)


def _cmd_info(blk):
    for name, value in sorted(blk.get_conf().items()):
        print(f"{name}: {value}")


def _cmd_list():
    circuit = get_circuit()
    for name, btype in sorted((blk.name, type(blk).__name__) for blk in circuit.getblocks()):
        print(f"{name}   ({btype})")


def _cmd_put(blk, value):
    retval = blk.put(ast.literal_eval(value))
    print(f"put() returned: {retval}")
    _cmd_show(blk)


def _cmd_show(blk):
    output = blk.output
    state = blk.get_state()
    if state == output:
        print(f"output and state: {output!r}")
    else:
        print(f"output: {output!r}")
        print(f"state: {state!r}")


def _complete(cmd):
    if cmd == '?':
        return 'help'
    for fullcmd in _PARSE:
        # no abbreviation for eval
        if fullcmd.startswith(cmd) and not 'eval' == fullcmd != cmd:
            return fullcmd
    raise ValueError(f"Unknown command: {cmd}, try: help")


@dataclass(frozen=True)
class CmdInfo:
    func: Callable
    blk: bool = True    # first arg is block name
    args: int = 0       # number of mandatory arguments (excl. the block name if blk is set)
    optargs: int = 0    # number of optional arguments

_PARSE = {
    'cdebug': CmdInfo(func=_cmd_cdebug, blk=False, args=1),
    'debug': CmdInfo(func=_cmd_debug, args=1),
    'eval': CmdInfo(func=_cmd_eval, blk=False, args=1),
    'event': CmdInfo(func=_cmd_event, args=1, optargs=1),
    # exit is handled in the loop
    'help': CmdInfo(func=_cmd_help, blk=False),
    'info': CmdInfo(func=_cmd_info),
    'list': CmdInfo(func=_cmd_list, blk=False),
    'put': CmdInfo(func=_cmd_put, args=1),
    'show': CmdInfo(func=_cmd_show),
}


async def repl():
    """REPL = read, evaluate, print loop."""
    loop = asyncio.get_event_loop()
    rstream = asyncio.StreamReader(limit=LIMIT, loop=loop)
    protocol = asyncio.StreamReaderProtocol(rstream, loop=loop)
    await loop.connect_read_pipe(lambda: protocol, sys.stdin)
    # assuming stdout is not blocking to keep this demo simple

    history = [None] * 10
    circuit = get_circuit()
    idx = 0
    print("Type 'help' to get a summary of available commands.")
    while True:
        await asyncio.sleep(0)
        print(f'--- edzed {idx}> ', end='', flush=True)
        line = (await rstream.readline()).decode()
        if not line:
            print('received EOF')
            break
        line = line.strip()
        if not line:
            continue
        cmd, *args = line.split(None, 1)
        if cmd == 'exit':
            # arguments ignored
            break
        try:
            if len(cmd) == 2 and cmd[0] == '!' and cmd[1] in "0123456789?!":
                # history
                if args:
                    raise TypeError(f"{cmd} command takes no arguments")
                hcmd = cmd[1]
                prev = (idx - 1) % 10
                if hcmd == '?':
                    for i, hist in enumerate(history):
                        if hist is not None:
                            print(f"{'!' if i == prev else ' '}{i}> {hist[0]}")
                else:
                    # !0-!9 or !!
                    hist = history[int(hcmd) if '0' <= hcmd <= '9' else prev]
                    if hist is None:
                        raise LookupError("history: command not found")
                    cmd, func, args = hist
                    print(f"--- edzed history> {cmd}")
                    func(*args)
            else:
                cmd = _complete(cmd)
                cmdinfo = _PARSE[cmd]
                minargs = cmdinfo.args + (1 if cmdinfo.blk else 0)
                maxargs = minargs + cmdinfo.optargs
                if args != []:
                    args = args[0].split(None, maxargs - 1)
                if not minargs <= len(args) <= maxargs:
                    if minargs == maxargs:
                        raise TypeError(f"{cmd} command takes {minargs} argument(s)")
                    raise TypeError(f"{cmd} command takes {minargs} to {maxargs} arguments")
                if cmdinfo.blk:
                    args[0] = circuit.findblock(args[0])
                cmdinfo.func(*args)
                history[idx] = (line, cmdinfo.func, args)
                idx = (idx + 1) % 10
        except Exception as err:
            print(f"ERROR: {err}")


async def run_demo():
    """Run a circuit simulation and an edzed REPL."""
    logging.basicConfig(level=logging.INFO)
    tsim = asyncio.create_task(get_circuit().run_forever())
    trepl = asyncio.create_task(repl())
    await asyncio.wait([tsim, trepl], return_when=asyncio.FIRST_COMPLETED)
    for task in (tsim, trepl):
        if not task.done():
            task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass
        except Exception as err:
            print(f"Error: {err!r}")



"""
Event filters.

- - - - - -
Docs: https://edzed.readthedocs.io/en/latest/
Home: https://github.com/xitop/edzed/
"""

import logging

from .. import block


__all__ = ['not_from_undef', 'Edge', 'Delta', 'DataEdit']

_logger = logging.getLogger(__package__)


def not_from_undef(data):
    """Filter out the initial change from UNDEF to the first real value."""
    return data.get('previous', block.UNDEF) is not block.UNDEF


class Edge:
    """
    Event filter for logical values.
    """

    def __init__(self, rise=False, fall=False, u_rise=None, u_fall=False):
        self._rise = bool(rise)
        self._fall = bool(fall)
        self._urise = bool(u_rise) if u_rise is not None else self._rise
        self._ufall = bool(u_fall)
        if not (rise or fall or u_rise or u_fall):
            _logger.warning(
                "%s: all events will be filtered out!",
                type(self).__name__)

    def __call__(self, data):
        value = data['value']
        previous = data['previous']
        if previous is block.UNDEF:
            if self._urise if value else self._ufall:
                return True
        else:
            previous = bool(previous)
            if (not previous and self._rise) if value else (previous and self._fall):
                return True
        return False


class Delta:
    """
    Event filter for numeric values.
    """

    def __init__(self, delta):
        self._delta = delta
        self._last = block.UNDEF

    def __call__(self, data):
        value = data['value']
        if self._last is block.UNDEF or abs(self._last - value) >= self._delta:
            self._last = value
            return True
        return False


class DataEdit:
    """
    Modify the event data.
    """

    def __init__(self, func):
        self._func = func

    def __getattribute__(self, name):
        """
        Disallow chaining of edit functions.

        Without this check, expressions similar to this one:
            DataEdit.add(key1).delete(key2)
        would be valid code, but wouldn't work as expected. Only the
        last operation will be performed. We better disallow this
        type of usage entirely.
        """
        attr = super().__getattribute__(name)   # raises if attr not found
        if not name.startswith('_'):
            # hide the filter functions in instances, the class must be the only source
            raise AttributeError(
                f"'{type(self).__name__}' object does not allow access to attribute '{name}'")
        return attr

    @classmethod
    def add(cls, **kwargs):
        """Add key=value pairs. Existing values wil be overwritten."""
        return cls(lambda data: {**data, **kwargs})

    @classmethod
    def default(cls, **kwargs):
        """Add key=value pairs only if key is missing."""
        return cls(lambda data: {**kwargs, **data})

    @classmethod
    def delete(cls, *args):
        """Delete listed keys. Non-existing keys are ignored."""
        def _edit(data):
            for key in args:
                data.pop(key, None)
            return data
        return cls(_edit)

    @classmethod
    def permit(cls, *args):
        """Delete all but listed keys."""
        def _edit(data):
            for key in list(data):
                if key not in args:
                    del data[key]
            return data
        return cls(_edit)

    @classmethod
    def copy(cls, src, dst):
        """Copy data[src] to data[dst]."""
        def _edit(data):
            data[dst] = data[src]
            return data
        return cls(_edit)

    REJECT = object()

    @classmethod
    def modify(cls, key, func):
        """Apply the func to a value identified by key."""
        def _edit(data):
            current = data[key]
            replacement = func(current)
            if replacement is cls.REJECT:
                return None
            data[key] = replacement
            return data
        return cls(_edit)

    def __call__(self, data):
        return self._func(data)
